import numpy as np
from numba.typed import List
from typing import Tuple, Optional, Union
from enum import Enum
from scipy.integrate import solve_ivp

from stochvolmodels.pricers.factor_hjm.rate_factor_basis import NelsonSiegel
from stochvolmodels.pricers.factor_hjm.rate_core import bracket, divide_mc, prod_mc, get_futures_start_and_pmt
from stochvolmodels.pricers.factor_hjm.rate_evaluate import swap_rate, annuity, bond
from stochvolmodels.pricers.factor_hjm.rate_affine_expansion import compute_logsv_a_mgf_grid, UnderlyingType
from stochvolmodels.pricers.factor_hjm.rate_logsv_params import RateLogSvParams, pw_const, get_default_swap_term_structure, MultiFactRateLogSvParams
from stochvolmodels.pricers.analytic.bachelier import infer_normal_implied_vol, infer_normal_ivols_from_slice_prices
from stochvolmodels.pricers.logsv.affine_expansion import ExpansionOrder, get_expansion_n
from stochvolmodels.pricers.model_pricer import ModelPricer, ModelParams
from stochvolmodels.data.option_chain import OptionChain, SwOptionChain, FutOptionChain
from stochvolmodels.utils.funcs import to_flat_np_array, set_time_grid, timer


from stochvolmodels.pricers.factor_hjm.double_exp_pricer import de_pricer

class Measure(Enum):
    RISK_NEUTRAL = 1
    ANNUITY = 2
    FORWARD = 3


class FutSettleType(Enum):
    EURODOLLAR = 1
    SOFR = 2


def conv_adj_rhs_MF(tau: float,
                    state0: np.ndarray,
                    ttm: float,
                    params: MultiFactRateLogSvParams,
                    Delta: float,
                    settlement_type: FutSettleType,
                    expansion_order: ExpansionOrder
                    ) -> np.ndarray:
    M = params.M
    C = params.C
    Omega = params.Omega
    sigma0 = params.sigma0
    q = params.theta if params.q is None else params.q
    ts = params.beta.ts
    betaxs = params.beta.xs
    volvolxs = params.volvol.xs

    sz_X = params.basis.nb_factors
    sz_Y = params.basis.nb_aux_factors
    D_X = params.basis.get_generating_matrix()
    D_Y = params.basis.get_aux_generating_matrix()

    if expansion_order == ExpansionOrder.FIRST:
        sz_vol = 3
        if settlement_type == FutSettleType.SOFR:
            raise NotImplementedError
    elif expansion_order == ExpansionOrder.ZERO:
        sz_vol = 2
    else:
        raise NotImplementedError

    sz = sz_X + sz_Y + sz_vol
    assert state0.size == sz
    B1, B2 = state0[:sz_X], state0[sz_X:sz_X + sz_Y]
    if expansion_order == ExpansionOrder.FIRST:
        h1, h2, h0 = state0[-3], state0[-2], state0[-1]
    elif expansion_order == ExpansionOrder.ZERO:
        h1, h0 = state0[-2], state0[-1]
        h2 = 0.0
    else:
        raise NotImplementedError

    v0 = sigma0 - q

    idx_t = bracket(ts[1:], ttm - tau, True)
    kappa0 = params.kappa1 * (params.theta - q) + params.kappa2 * q * (params.theta - q)
    kappa1 = params.kappa1 - params.kappa2 * params.theta + 2.0*params.kappa2 * q
    kappa2 = params.kappa2
    beta_t = betaxs[idx_t]
    volvol_t = volvolxs[idx_t]
    vartheta_sq = np.linalg.norm(beta_t) ** 2 + volvol_t ** 2
    C_t = C[idx_t]
    M_t = M[idx_t]
    Omega_t = Omega[idx_t]

    B0 = params.basis.get_basis(0.0)
    B0_ext = params.basis.get_aux_basis(0.0)

    # interim variables
    B1_M_B1 = np.dot(np.dot(np.transpose(B1), M_t), B1)
    B1_C_beta = np.dot(np.dot(B1, C_t), beta_t)
    B2_Omega = np.dot(B2, Omega_t)

    rhs = np.zeros_like(state0)
    rhs[:sz_X] = np.dot(B1, D_X)
    rhs[sz_X:sz_X + sz_Y] = np.dot(B2, D_Y)
    if settlement_type == FutSettleType.SOFR:
        rhs[:sz_X] += B0 if tau < Delta else 0.0
        rhs[sz_X:sz_X + sz_Y] += B0_ext if tau < Delta else 0.0

    if expansion_order == ExpansionOrder.FIRST:
        # h1
        rhs[-3] = 2.0 * q * (0.5 * B1_M_B1 + B2_Omega) + 2.0 * kappa0 * h2 - kappa1 * h1 + vartheta_sq * (
                q * h1 * h1 + 2.0 * q * h2 + 2.0 * q * q * h1 * h2) + 2.0 * q * (h1 + q * h2) * B1_C_beta
        # h2
        rhs[-2] = (0.5 * B1_M_B1 + B2_Omega) - 2.0 * kappa1 * h2 - kappa2 * h1 + vartheta_sq * (
                0.5 * h1 * h1 + h2 + 4.0 * q * h1 * h2 + 2.0 * q * q * h2 * h2) + (h1 + 4.0 * q * h2) * B1_C_beta
        # h0
        rhs[-1] = q * q * (0.5 * B1_M_B1 + B2_Omega) + kappa0 * h1 + vartheta_sq * q * q * (
                0.5 * h1 * h1 + h2) + q * q * h1 * B1_C_beta
    elif expansion_order == ExpansionOrder.ZERO:
        # h1
        rhs[-2] = 2.0*q * (0.5*B1_M_B1 + B2_Omega + h1*B1_C_beta + 0.5*vartheta_sq*h1*h1) - kappa1 * h1
        # h0
        rhs[-1] = q * q * (0.5*B1_M_B1 + B2_Omega + h1*B1_C_beta + 0.5*vartheta_sq*h1*h1) + kappa0 * h1


    return rhs


def futures_conv_adj(t_start: float,
                     basis_type: str,
                     params: MultiFactRateLogSvParams,
                     t0: float,
                     Delta: float,
                     settlement_type: FutSettleType,
                     expansion_order: ExpansionOrder,
                     dense_output: bool = False,
                     t_grid: np.ndarray = None) -> Tuple[np.ndarray, ...]:
    assert basis_type == "NELSON-SIEGEL"

    bond_coeffs = params.basis.bond_coeffs(Delta)
    if expansion_order == ExpansionOrder.FIRST:
        vol_init = np.zeros((3,))
    elif expansion_order == ExpansionOrder.ZERO:
        vol_init = np.zeros((2,))
    else:
        raise NotImplementedError

    if settlement_type == FutSettleType.EURODOLLAR:
        cond_init = np.concatenate((bond_coeffs[0],
                                    bond_coeffs[1],
                                    vol_init))
    elif settlement_type == FutSettleType.SOFR:
        cond_init = np.concatenate((np.zeros_like(bond_coeffs[0]),
                                    np.zeros_like(bond_coeffs[1]),
                                    vol_init))
    else:
        raise NotImplementedError

    assert t0 <= t_start  # check for SOFR options
    tau_S = t_start - t0
    tau_E = tau_S + Delta
    t_eval = np.maximum(t_start-t0, 1e-4) if settlement_type == FutSettleType.EURODOLLAR else np.maximum(t_start+Delta-t0, 1e-4)

    if t_grid is not None:
        idx_ttm = np.where(t_grid == t_eval)[0][0]
        t_grid = t_grid[:idx_ttm + 1]

    # t_grid = np.arange(0.0, t_eval+1e-6, 0.025)

    sol_futconvex = solve_ivp(fun=conv_adj_rhs_MF, y0=cond_init,
                              args=(t_start, params, Delta, settlement_type, expansion_order),
                              t_span=(0.0, t_eval), dense_output=dense_output,
                              t_eval=t_grid, max_step=0.001)
    sol = sol_futconvex.y[:, -1]
    sz_X = params.basis.nb_factors
    sz_Y = params.basis.nb_aux_factors

    if expansion_order == ExpansionOrder.FIRST:
        b1, b2, h1, h2, h0 = sol[:sz_X], sol[sz_X:sz_X + sz_Y], sol[-3], sol[-2], sol[-1]
    elif expansion_order == ExpansionOrder.ZERO:
        b1, b2, h1, h0 = sol[:sz_X], sol[sz_X:sz_X + sz_Y], sol[-2], sol[-1]
        h2 = 0.0
    # now we take into account exp[-(...)X_t-(...)Y_t]
    # need to convert from tuple to list to support arithmetic operations
    b1 = b1 - (params.basis.bond_coeffs(tau_E)[0] - params.basis.bond_coeffs(tau_S)[0])
    b2 = b2 - (params.basis.bond_coeffs(tau_E)[1] - params.basis.bond_coeffs(tau_S)[1])

    if dense_output:
        assert t_grid is not None
        # we solve ODE in time-to-maturity tau=T-t, but need to return it in time t
        sol = sol_futconvex.sol(t_start-t_grid)
        b1 = sol.T[:, :sz_X]
        b2 = sol.T[:, sz_X:sz_X + sz_Y]
        h1 = sol.T[:, -2]
        h0 = sol.T[:, -1]
        if expansion_order == ExpansionOrder.FIRST:
            h2 = sol.T[:, -3]
        elif expansion_order == ExpansionOrder.ZERO:
            h2 = np.zeros_like(t_grid)

    return b1, b2, h1, h2, h0


def calc_futures_rate(ccy: str,
                      basis_type: str,
                      params: MultiFactRateLogSvParams,
                      x0: np.ndarray,
                      y0: np.ndarray,
                      sigma0: np.ndarray,
                      t0: float,
                      t_start: float,
                      t_end: float,
                      Delta: float,
                      settlement_type: FutSettleType,
                      expansion_order: ExpansionOrder) -> Tuple[np.ndarray, ...]:
    assert basis_type == "NELSON-SIEGEL"
    assert 0 <= t0 <= t_start
    q = params.theta if params.q is None else params.q
    v0 = sigma0[:, 0] - q
    b1, b2, h1, h2, h0 = futures_conv_adj(t_start=t_start,
                                          basis_type=basis_type,
                                          params=params,
                                          t0=t0,
                                          Delta=Delta,
                                          settlement_type=settlement_type,
                                          expansion_order=expansion_order)
    c_tau = np.exp(np.dot(b1, np.transpose(x0)) + np.dot(b2, np.transpose(y0)) + h0 + h1 * v0 + h2 * v0 * v0)
    P_t_Ts_Te = params.basis.bond(t=t0, T=t_end, x=x0, y=y0, ccy=ccy, m=0) / params.basis.bond(t=t0, T=t_start, x=x0,
                                                                                               y=y0, ccy=ccy, m=0)
    if x0 is None:
        x0 = np.zeros((params.basis.get_nb_factors(),))
    else:
        assert x0.ndim == 1 and x0.shape[0] == params.basis.get_nb_factors()
    if y0 is None:
        y0 = np.zeros((params.basis.get_nb_aux_factors(),))
    else:
        assert y0.ndim == 1 and y0.shape[0] == params.basis.get_nb_aux_factors()
    P_0_Ts_Te = params.basis.bond(t=t0, T=t_end, x=x0, y=y0, ccy=ccy, m=0)[0] / params.basis.bond(t=t0, T=t_start, x=x0, y=y0, ccy=ccy, m=0)[0]


    # futures convexity adjustment esimates E[exp(B_P*(X_T-X_t)+...)] thus we multiply by P(t,T_S,T_E)
    futures_analyt_ae1 = 1.0 / Delta * (1.0 / P_t_Ts_Te * c_tau - 1.0)

    return futures_analyt_ae1, c_tau, P_t_Ts_Te, P_0_Ts_Te


def logsv_chain_de_pricer(params: MultiFactRateLogSvParams,
                          t_grid: np.ndarray,
                          ttms: np.ndarray,
                          forwards: List[np.ndarray],
                          strikes_ttms: List[List[np.ndarray]],
                          optiontypes_ttms: List[np.ndarray],
                          is_stiff_solver: bool = False,
                          underlying_type: UnderlyingType = UnderlyingType.SWAP,
                          is_analytic: bool = False,
                          is_spot_measure: bool = True,
                          expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
                          vol_scaler: float = None,
                          do_control_variate=False,
                          x0: np.ndarray = None,
                          y0: np.ndarray = None,
                          **kwargs) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    model_prices_tenors_ttms = List()
    model_ivs_tenors_ttms = List()
    t_grid0 = t_grid
    if 'lag' in kwargs:
        lag = kwargs['lag']
    else:
        lag = 0.0
    if 'settlement_type' in kwargs:
        settlement_type = kwargs['settlement_type']
    else:
        settlement_type = FutSettleType.EURODOLLAR
    if underlying_type == UnderlyingType.SWAP:
        # sanity check: we must have as much forwards as tenors for swaptions
        assert params.basis.key_terms.size == len(forwards)
        # we calibrate expiry by expiry
        # per expiry we calculate implied vols tenor-by-tenor
        assert ttms.size == 1 and len(optiontypes_ttms) == 1
        ttms_ = np.ones_like(params.basis.key_terms) * ttms[0]
        optiontypes_ttms_ = [optiontypes_ttms[0] for tenor in params.basis.key_terms]
        rng_ttm = params.basis.key_terms

    elif underlying_type == UnderlyingType.FUTURES:
        assert len(forwards) == 1
        # we calibrate expiry by expiry
        assert ttms.size == 1 and len(optiontypes_ttms) == 1
        ttms_ = ttms
        optiontypes_ttms_ = optiontypes_ttms
        rng_ttm = ['FUTURES_DUMMY_TENOR']
    else:
        raise NotImplementedError

    for idx_tenor, _ in enumerate(rng_ttm):
        model_prices_ttms = List()
        model_ivs_ttms = List()

        tenor = np.nan
        if underlying_type == UnderlyingType.SWAP:
            tenor = rng_ttm[idx_tenor]

        for ttm, forward, strikes_ttm, optiontypes_ttm in zip(ttms_, forwards[idx_tenor], strikes_ttms[idx_tenor],
                                                              optiontypes_ttms_):
            if underlying_type == UnderlyingType.SWAP:
                a, kappa0, kappa1, kappa2, beta, volvol, _ = params.transform_QA_params(expiry=ttm, t_grid=t_grid0,
                                                                                        tenor=tenor, x0=x0, y0=y0)
                a0 = a
                a1 = np.zeros_like(kappa0)
                b = np.zeros_like(kappa0)

            elif underlying_type == UnderlyingType.FUTURES:
                tenor = 0.25
                start, end = get_futures_start_and_pmt(t0=ttm, lag=0.0, libor_tenor=tenor)
                frac = end - start
                a, eta, kappa0, kappa1, kappa2, beta, volvol = params.transform_QT_params(expiry=ttm, t_grid=t_grid,
                                                                                          t_start=start, t_end=end)

                b1, b2, h1, h2, h0 = futures_conv_adj(t_start=start,
                                                      basis_type="NELSON-SIEGEL",  # hard-coded
                                                      params=params,
                                                      t0=0.0,
                                                      Delta=tenor, expansion_order=ExpansionOrder.ZERO,
                                                      dense_output=True,
                                                      t_grid=t_grid, settlement_type=settlement_type)
                # for futures we have extra contribution
                a0 = a + np.einsum('i,ij->ij', h1, beta)
                a1 = np.multiply(h1, volvol)
                b = np.einsum('ij,ij->i', a0, eta) + 0.5 * np.einsum('ij,ij->i', a0, a0)
            else:
                NotImplementedError(f"Underlying type is not supported")
            # cut t_grid
            itemindex = np.where(t_grid0 == ttm)
            itemindex = itemindex[0][0]
            t_grid = t_grid0[:itemindex + 1]

            # delta = np.zeros_like(delta)  # TODO: remove. just for testing
            # params.q = 0.0  # TODO: q=0 just for testing. reset it for pricing
            def ff(p: np.ndarray) -> np.ndarray:
                real_phi = -0.5
                phi_grid = real_phi + 1j * p
                log_mgf_grid = compute_logsv_a_mgf_grid(ttm=ttm,
                                                        phi_grid=phi_grid,
                                                        sigma0=params.sigma0,
                                                        q=params.q,
                                                        times=t_grid,
                                                        a0=a0,
                                                        a1=a1,
                                                        kappa0=kappa0,
                                                        kappa1=kappa1,
                                                        kappa2=kappa2,
                                                        beta=beta,
                                                        volvol=volvol,
                                                        b=b,
                                                        underlying_type=underlying_type,
                                                        expansion_order=expansion_order,
                                                        is_stiff_solver=is_stiff_solver)[1]
                integrand_ = np.zeros((phi_grid.shape[0], strikes_ttm.shape[0]))
                if underlying_type == UnderlyingType.SWAP:
                    moneyness = strikes_ttm - forward
                    p_payoff = (1.0 / np.pi) / (phi_grid * phi_grid)
                    for idx, (x, strike, type_) in enumerate(zip(moneyness, strikes_ttm, optiontypes_ttm)):
                        integrand_[:, idx] = np.real(p_payoff * (np.exp(x*phi_grid + log_mgf_grid)))
                elif underlying_type == UnderlyingType.FUTURES:
                    moneyness = np.log((strikes_ttm + 1.0/frac)/(forward + 1.0/frac))
                    p_payoff = (1.0 / np.pi) / (phi_grid * (phi_grid + 1.0))
                    for idx, (x, strike, type_) in enumerate(zip(moneyness, strikes_ttm, list(optiontypes_ttm))):
                        integrand_[:, idx] = np.real(
                            p_payoff * (-(strike + 1.0 / frac)) * (np.exp(x * phi_grid + log_mgf_grid)))

                else:
                    NotImplementedError("Underlying type is not supported")



                np.set_printoptions(linewidth=np.inf, precision=16)
                return integrand_

            # in case of options on swap rate, we directly price call options
            # in case of options on futures, we price capped payoff and call option through capped payoff
            if underlying_type == UnderlyingType.SWAP:
                def ff_transf(model_prices: np.ndarray):
                    normal_ivols = infer_normal_ivols_from_slice_prices(ttm=ttm, forward=forward,
                                                                        strikes=strikes_ttm,
                                                                        model_prices=model_prices[0,:],
                                                                        optiontypes=np.repeat('C', strikes_ttm.size),
                                                                        discfactor=1.0)
                    return model_prices, normal_ivols
            elif underlying_type == UnderlyingType.FUTURES:
                def ff_transf(capped_prices: np.ndarray) -> (np.ndarray, np.ndarray):
                    call_prices = forward + 1.0/frac - capped_prices
                    normal_ivols = infer_normal_ivols_from_slice_prices(ttm=ttm, forward=forward, strikes=strikes_ttm,
                                                                        model_prices=call_prices[0, :],
                                                                        optiontypes=np.repeat('C', strikes_ttm.size),
                                                                        discfactor=1.0)
                    return call_prices, normal_ivols
            else:
                NotImplementedError(f"underlying type is unknown")


            model_prices_ttm, model_ivs_ttm = de_pricer(ff, ff_transf)
            model_prices_ttms.append(model_prices_ttm[0, :])
            model_ivs_ttms.append(model_ivs_ttm)
        model_prices_tenors_ttms.append(model_prices_ttms)
        model_ivs_tenors_ttms.append(model_ivs_ttms)
    return model_prices_tenors_ttms, model_ivs_tenors_ttms


class RateLogSVPricer(ModelPricer):
    def price_chain(self,
                    option_chain: SwOptionChain,
                    params: Union[RateLogSvParams, MultiFactRateLogSvParams],
                    is_spot_measure: bool = True,
                    **kwargs
                    ) -> List[np.ndarray]:
        """
        implementation of generic method price_chain using log sv wrapper
        """
        t_grid = kwargs['t_grid']
        idxs = kwargs['idxs']
        ttms = np.array(option_chain.ttms[idxs])
        forwards = [option_chain.forwards[idx_tenor][idxs] for idx_tenor, _ in enumerate(option_chain.tenors)]
        strikes_ttms = [option_chain.strikes_ttms[idx_tenor][idxs] for idx_tenor, _ in enumerate(option_chain.tenors)]
        optiontypes_ttms = option_chain.optiontypes_ttms[idxs]

        model_ivols = logsv_chain_de_pricer(params=params,
                                            t_grid=t_grid,
                                            ttms=ttms,
                                            forwards=forwards,
                                            strikes_ttms=strikes_ttms,
                                            optiontypes_ttms=optiontypes_ttms,
                                            expansion_order=ExpansionOrder.FIRST,
                                            is_stiff_solver=False)[1]
        return model_ivols

    def model_mc_price_chain(self,
                             option_chain: SwOptionChain,
                             params: RateLogSvParams,
                             nb_path: int = 100000,
                             **kwargs
                             ) -> (List[np.ndarray], List[np.ndarray]):
        return logsv_mc_chain_pricer(sigma0=params.sigma0,
                                     theta=params.theta,
                                     kappa1=params.kappa1,
                                     kappa2=params.kappa2,
                                     ts=params.alpha.ts,
                                     betaxs=params.beta.xs,
                                     volvolxs=params.volvol.xs,
                                     ttms=option_chain.ttms,
                                     forwards=option_chain.forwards,
                                     strikes_ttms=option_chain.strikes_ttms,
                                     optiontypes_ttms=option_chain.optiontypes_ttms,
                                     nb_path=nb_path,
                                     **kwargs)


class RateFutLogSVPricer(ModelPricer):
    def price_chain(self,
                    option_chain: FutOptionChain,
                    params: MultiFactRateLogSvParams,
                    is_spot_measure: bool = True,
                    **kwargs
                    ) -> List[np.ndarray]:
        """
        implementation of generic method price_chain using log sv wrapper
        """
        t_grid = kwargs['t_grid']
        idxs = kwargs['idxs']
        ttms = np.array(option_chain.ttms[idxs])
        forwards = [option_chain.forwards[idxs]]
        strikes_ttms = [option_chain.strikes_ttms[idxs]]
        optiontypes_ttms = [option_chain.optiontypes_ttms[0]]
        if 'expansion_order' in kwargs:
            expansion_order = kwargs['expansion_order']
        else:
            expansion_order = ExpansionOrder.FIRST
        if 'is_stiff_solver' in kwargs:
            is_stiff_solver = kwargs['is_stiff_solver']
        else:
            is_stiff_solver = True
        if 'x0' in kwargs:
            x0 = kwargs['x0']
        else:
            x0 = None
        if 'y0' in kwargs:
            y0 = kwargs['y0']
        else:
            y0 = None


        model_ivols = logsv_chain_de_pricer(params=params,
                                            t_grid=t_grid,
                                            ttms=ttms,
                                            forwards=forwards,
                                            strikes_ttms=strikes_ttms,
                                            optiontypes_ttms=optiontypes_ttms,
                                            underlying_type=UnderlyingType.FUTURES,
                                            lag=0, # TODO: review it for mid-curve options
                                            expansion_order=expansion_order,
                                            is_stiff_solver=is_stiff_solver,
                                            x0=x0,
                                            y0=y0)[1]
        return model_ivols

    @classmethod
    def populate_betas(cls, beta: float, basis: NelsonSiegel) -> np.ndarray:
        if basis.get_nb_factors() == 3:
            return np.array([beta, -0.5*beta, 0.0])
        elif basis.get_nb_factors() == 1:
            return np.array([beta])
        else:
            raise NotImplementedError


def update_params(param0: MultiFactRateLogSvParams,
                  idx: int,
                  opt_val: np.ndarray):
    nb_factors = param0.basis.get_nb_factors()
    a_ttm, beta_ttm, volvol_ttm = opt_val[:nb_factors] * 0.01, opt_val[nb_factors:nb_factors + nb_factors], opt_val[-1]
    # update parameters
    param0.update_params(idx=idx, A_idx=a_ttm, beta_idx=beta_ttm, volvol_idx=volvol_ttm)


def logsv_mc_chain_pricer(sigma0: float,
                          theta: float,
                          kappa1: float,
                          kappa2: float,
                          ts: np.ndarray,
                          axs: np.ndarray,
                          bxs: np.ndarray,
                          betaxs: np.ndarray,
                          volvolxs: np.ndarray,
                          ttms: np.ndarray,
                          forwards: np.ndarray,
                          strikes_ttms: List[np.ndarray],
                          optiontypes_ttms: List[np.ndarray],
                          lamda: np.ndarray = None,
                          is_annuity_measure: bool = False,
                          is_approx_dynamics: bool = False,
                          nb_path: int = 100000,
                          seed: int = None,
                          ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    # starting values
    sigma0 = sigma0 * np.ones(nb_path)

    # outputs as numpy lists
    option_prices_ttm = List()
    option_std_ttm = List()
    for ttm, forward, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, strikes_ttms, optiontypes_ttms):
        x0 = np.zeros((nb_path,))
        y0 = np.zeros((nb_path,))
        I0 = np.zeros((nb_path,))
        if is_approx_dynamics:
            s0 = simulate_logsv_swap_approx_terminal(ttm=ttm,
                                                     sigma0=sigma0,
                                                     theta=theta,
                                                     kappa1=kappa1,
                                                     kappa2=kappa2,
                                                     ts=ts,
                                                     axs=axs,
                                                     betaxs=betaxs,
                                                     volvolxs=volvolxs,
                                                     seed=seed,
                                                     nb_path=nb_path)
            option_prices, option_std = compute_mcapprox_payoff(ttm=ttm,
                                                                s_mc=s0,
                                                                strikes_ttm=strikes_ttm,
                                                                optiontypes_ttm=optiontypes_ttm)
        else:
            x0, y0, I0, qvar0, _ = simulate_logsv_x_vol_terminal(ttm=ttm,
                                                                 x0=x0,
                                                                 y0=y0,
                                                                 I0=I0,
                                                                 sigma0=sigma0,
                                                                 theta=theta,
                                                                 kappa1=kappa1,
                                                                 kappa2=kappa2,
                                                                 ts=ts,
                                                                 axs=axs,
                                                                 bxs=bxs,
                                                                 betaxs=betaxs,
                                                                 volvolxs=volvolxs,
                                                                 lamda=lamda,
                                                                 nb_path=nb_path,
                                                                 seed=seed,
                                                                 is_annuity_measure=is_annuity_measure)

            # np.savetxt('c:/temp/dump.txt', qvar0)

            option_prices, option_std = compute_mc_vars_payoff(ttm=ttm,
                                                               x0=x0,
                                                               y0=y0,
                                                               I0=I0,
                                                               strikes_ttm=strikes_ttm,
                                                               optiontypes_ttm=optiontypes_ttm,
                                                               is_annuity_measure=is_annuity_measure)
        option_prices_ttm.append(option_prices)
        option_std_ttm.append(option_std)

    return option_prices_ttm, option_std_ttm


# @njit(cache=False, fastmath=True)  # TODO
def simulate_logsv_x_vol_terminal(ttm: float,
                                  x0: np.ndarray,
                                  y0: np.ndarray,
                                  I0: np.ndarray,
                                  sigma0: np.ndarray,
                                  theta: float,
                                  kappa1: float,
                                  kappa2: float,
                                  ts: np.ndarray,
                                  axs: np.ndarray,
                                  bxs: np.ndarray,
                                  betaxs: np.ndarray,
                                  volvolxs: np.ndarray,
                                  lamda: np.ndarray = None,
                                  is_heston: bool = False,
                                  is_annuity_measure: bool = False,
                                  nb_path: int = 100000,
                                  qvar0: np.ndarray = None,
                                  seed: int = None
                                  ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    ttms = np.array([ttm])
    res = simulate_logsv_x_vol(ttms=ttms, x0=x0, y0=y0, I0=I0, sigma0=sigma0, theta=theta, kappa1=kappa1, kappa2=kappa2,
                               ts=ts, axs=axs, bxs=bxs, betaxs=betaxs, volvolxs=volvolxs, lamda=lamda,
                               is_heston=is_heston,
                               is_annuity_measure=is_annuity_measure,
                               nb_path=nb_path,
                               qvar0=qvar0,
                               seed=seed)
    x0 = res[0][-1]
    y0 = res[1][-1]
    I0 = res[2][-1]
    qvar0 = res[3][-1]
    sigma0 = res[4][-1]

    return x0, y0, I0, qvar0, sigma0


# @njit(cache=False, fastmath=True)  # TODO
def simulate_logsv_x_vol(ttms: np.ndarray,
                         x0: np.ndarray,
                         y0: np.ndarray,
                         I0: np.ndarray,
                         sigma0: np.ndarray,
                         theta: float,
                         kappa1: float,
                         kappa2: float,
                         ts: np.ndarray,
                         axs: np.ndarray,
                         bxs: np.ndarray,
                         betaxs: np.ndarray,
                         volvolxs: np.ndarray,
                         basis, # : Cheyette1D,
                         ts_sw: np.ndarray,
                         ccy: str,
                         is_heston: bool = False,
                         is_annuity_measure: bool = False,
                         nb_path: int = 100000,
                         seed: int = None,
                         cap_lsv: bool = False,
                         W: List[np.ndarray] = None,
                         ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    assert ttms.shape[0] > 0
    ttm = ttms[-1]
    if x0.shape[0] == 1:  # initial value
        x0 = x0 * np.zeros(nb_path)
    else:
        assert x0.shape[0] == nb_path

    if y0.shape[0] == 1:  # initial value
        y0 = y0 * np.zeros(nb_path)
    else:
        assert y0.shape[0] == nb_path

    if I0.shape[0] == 1:  # initial value
        I0 = I0 * np.zeros(nb_path)
    else:
        assert I0.shape[0] == nb_path

    if sigma0.shape[0] == 1:
        sigma0 = sigma0 * np.ones(nb_path)
    else:
        assert sigma0.shape[0] == nb_path

    # pre-generate random numbers
    if seed is None:
        seed = 16
    np.random.seed(seed)  # fix seed
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm)
    if W is None:
        W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))  # TODO: undo
        W1 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))  # TODO: undo
    else:
        W0 = W[0] * np.sqrt(dt)
        W1 = W[1] * np.sqrt(dt)
    # W0 = np.transpose(np.loadtxt('C:/temp/rnd1.txt', delimiter=',')) * np.sqrt(dt)
    # W1 = np.transpose(np.loadtxt('C:/temp/rnd2.txt', delimiter=',')) * np.sqrt(dt)

    idx_ttms = [np.where(grid_t == t)[0][0] for t in ttms]
    # print(idx_ttms)
    x0s = List()
    y0s = List()
    I0s = List()
    qvar0s = List()
    sigma0s = List()

    mrv_r = basis.meanrev
    # ts_sw = get_default_swap_term_structure(ttm)  # need when we simulate under Q^A
    sigma0_cpy = sigma0

    if 0 in idx_ttms:
        x0s.append(x0), y0s.append(y0), I0s.append(I0), sigma0s.append(sigma0)

    if not is_heston:
        log_vol = np.log(sigma0)

        for idx, (t_, w0, w1) in enumerate(zip(grid_t, W0, W1)):
            # reshape (n,) array to (n,1)
            w0 = np.reshape(w0, (nb_path, 1))
            w1 = np.reshape(w1, (nb_path, 1))

            a_t = pw_const(ts, axs, t_, flat_extrapol=False, shift=1)
            b_t = pw_const(ts, bxs, t_, flat_extrapol=False, shift=1)
            beta_t = pw_const(ts, betaxs, t_, flat_extrapol=False, shift=1)
            volvol_t = pw_const(ts, volvolxs, t_, flat_extrapol=False, shift=1)
            vartheta2 = beta_t * beta_t + volvol_t * volvol_t

            if is_annuity_measure:
                ann_der0_mc = basis.annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, m=0, ccy=ccy)
                ann_der1_mc = basis.annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, m=1, ccy=ccy)
                d_loga_dx = divide_mc(ann_der1_mc, ann_der0_mc)
                adj_x_drift = (a_t + b_t * x0) ** 2 * sigma0 ** 2 * d_loga_dx
                adj_vol_drift = (a_t + b_t * x0) * beta_t * sigma0 * d_loga_dx
            else:
                adj_x_drift = 0
                adj_vol_drift = 0

            I0 = I0 + x0[:,0] * dt
            # qvar0 = qvar0 + sigma0 * sigma0 * dt
            lsv = (a_t + b_t * x0) * sigma0
            lsv = np.minimum(lsv, 5.0) if cap_lsv else lsv  # cap to 500%
            y0 = y0 + (lsv ** 2 - 2.0 * mrv_r * y0) * dt
            x0 = x0 + (y0 - mrv_r * x0) * dt + lsv * w0 + adj_x_drift * dt
            log_vol = log_vol + ((kappa1 * theta / sigma0) - (
                    kappa1 - kappa2 * theta + 0.5 * vartheta2) - kappa2 * sigma0) * dt + beta_t * w0 + volvol_t * w1 + adj_vol_drift * dt
            sigma0 = np.exp(log_vol)
            # check that forwards are ok
            tau = 5.0
            mrv = mrv_r
            exp_mmt = np.exp(-mrv * tau)
            G_t_T = (1.0 - exp_mmt) / mrv
            fwd = exp_mmt * x0 + exp_mmt * G_t_T * y0
            # end of check
            if idx + 1 in idx_ttms:
                x0s.append(x0), y0s.append(y0), I0s.append(I0), sigma0s.append(sigma0), None

    else:  # case of heston dynamics
        # raise ValueError("Heston not implemented properly")
        if np.fabs(kappa2) >= 1e-8:
            raise ValueError("in heston model kappa2 must be zero")
        var0 = sigma0
        for idx, (t_, w0, w1) in enumerate(zip(grid_t, W0, W1)):
            vol0 = np.sqrt(var0)
            a_t = pw_const(ts, axs, t_, False)
            b_t = pw_const(ts, bxs, t_, False)
            beta_t = pw_const(ts, betaxs, t_, False)
            volvol_t = pw_const(ts, volvolxs, t_, False)

            if is_annuity_measure:
                ann_der0_mc = annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, m=0, is_mc_mode=True)
                ann_der1_mc = annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, m=1, is_mc_mode=True)
                d_loga_dx = ann_der1_mc / ann_der0_mc
                adj_x_drift = (a_t + b_t * x0) ** 2 * var0 * d_loga_dx
                adj_vol_drift = (a_t + b_t * x0) * beta_t * var0 * d_loga_dx
            else:
                adj_x_drift = 0
                adj_vol_drift = 0

            I0 = I0 + x0 * dt
            lsv = (a_t + b_t * x0) * vol0
            # capping with some useful diagnostics
            if not np.all(np.isfinite(lsv)):
                print(f"step = {idx}: explosion in Heston")
            if cap_lsv:
                cap = 5.0  # cap to 500%
                if np.any(lsv >= cap):
                    print(f"t={t_:.2f}: {np.argwhere(lsv >= cap).size} paths reached the cap: " + ''.join(
                        str(x) for x in np.argwhere(lsv >= cap)))
                lsv = np.minimum(lsv, cap)

            y0 = y0 + (lsv ** 2 - 2.0 * mrv_r * y0) * dt
            x0 = x0 + (y0 - mrv_r * x0) * dt + lsv * w0 + adj_x_drift * dt
            var0 = var0 + kappa1 * (theta - var0) * dt + vol0 * (beta_t * w0 + volvol_t * w1) + adj_vol_drift * dt
            var0 = np.maximum(var0, 1e-4)

            if idx + 1 in idx_ttms:
                x0s.append(x0), y0s.append(y0), I0s.append(I0)
                qvar0s.append(np.ones_like(x0)*np.nan)
                sigma0s.append(np.ones_like(x0)*np.nan)

    return x0s, y0s, I0s, qvar0s, sigma0s

def make_mc_array(x: np.ndarray, nb_path: int):
    x_ = np.zeros((nb_path, x.size))
    for idx, val in enumerate(x):
        x_[:, idx] = val
    return x_


def simulate_logsv_MF(ttms: np.ndarray,
                      x0: np.ndarray,
                      y0: np.ndarray,
                      I0: np.ndarray,
                      sigma0: np.ndarray,
                      theta: float,
                      kappa1: float,
                      kappa2: float,
                      ts: np.ndarray,
                      A: List[np.ndarray],  # for multi-factor model
                      R: List[np.ndarray],  # for multi-factor model
                      C: List[np.ndarray],  # for multi-factor model
                      Omega: List[np.ndarray],  # for multi-factor model
                      betaxs: np.ndarray,
                      volvolxs: np.ndarray,
                      basis: NelsonSiegel,
                      ts_sw: np.ndarray,
                      T_fwd: float,
                      ccy: str,
                      measure_type: Measure = Measure.RISK_NEUTRAL,
                      nb_path: int = 100000,
                      seed: int = None,
                      W: List[np.ndarray] = None,
                      bxs: np.ndarray = None,
                      params0: MultiFactRateLogSvParams = None,
                      year_days: int = 360,
                      quasi_random: bool = False
                      ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """Simulate factors X_t, Y_t, sigma_t, I_t for each t in ttms
    TODO: ideally we should not pass params0, but for DLN skew we need to recalculate volatility matrix C and
    TODO: auxiliary vector \\Omega_t for each simulation step, thus passed as optional parameter
    TODO: in potential numba implementation we should think of removing it
    """
    assert ttms.shape[0] > 0
    ttm = ttms[-1]

    assert ts.shape[0] > 0 and ts[0] == 0.0
    nb_times = ts.size - 1

    # check inputs
    nb_factors = basis.get_nb_factors()
    assert A.ndim == 2 and A.shape[0] == nb_times and A.shape[1] == nb_factors
    assert betaxs.ndim == 2 and betaxs.shape[0] == nb_times and betaxs.shape[1] == nb_factors
    assert volvolxs.ndim == 1 and volvolxs.shape[0] == nb_times
    assert R.ndim == 2 and R.shape[0] == R.shape[1] == nb_factors
    assert Omega.ndim == 2 and Omega.shape[0] == nb_times and Omega.shape[1] == basis.get_nb_aux_factors()
    # if b (displaced log-normal skew) is provided, we assume it is flat across expiries and values are per factor
    if bxs is not None:
        assert measure_type is Measure.RISK_NEUTRAL and np.all(np.fabs(betaxs) <= 1e-8) and np.all(volvolxs <= 1e-8)
        assert kappa1 <= 1e-8 and kappa2 <= 1e-8
        assert bxs.shape == (nb_factors,)

    if x0.shape[0] == basis.get_nb_factors():  # initial value
        x0 = make_mc_array(x0, nb_path)
    else:
        assert x0.shape == (nb_path, basis.get_nb_factors())

    if y0.shape[0] == basis.get_nb_aux_factors():  # initial value
        y0 = make_mc_array(y0, nb_path)
    else:
        assert y0.shape == (nb_path, basis.get_nb_aux_factors())

    if I0.shape[0] == 1:  # initial value
        I0 = I0 * np.zeros(nb_path)
    else:
        assert I0.shape[0] == nb_path

    if sigma0.shape[0] == 1:
        sigma0 = sigma0 * np.ones(nb_path)
    else:
        assert sigma0.shape[0] == nb_path

    # pre-generate random numbers
    if seed is None:
        seed = 16
    np.random.seed(seed)  # fix seed
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm, nb_steps_per_year=360)
    if W is None:
        W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path, basis.get_nb_factors()))  # TODO: undo
        W1 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))  # TODO: undo

    else:
        W0 = W[0] * np.sqrt(dt)
        W1 = W[1] * np.sqrt(dt)
    # W0 = np.transpose(np.loadtxt('C:/temp/rnd1.txt', delimiter=',')) * np.sqrt(dt)
    # W1 = np.transpose(np.loadtxt('C:/temp/rnd2.txt', delimiter=',')) * np.sqrt(dt)

    idx_ttms = [np.where(np.isclose(grid_t, t))[0][0] for t in ttms]
    x0s = []
    y0s = []
    I0s = []
    sigma0s = []

    if 0 in idx_ttms:
        x0s.append(x0), y0s.append(y0), I0s.append(I0), sigma0s.append(sigma0)

    log_vol = np.log(sigma0)
    D_X = basis.get_generating_matrix()
    D_Y = basis.get_aux_generating_matrix()
    # for simulation under Q we need short rate r(t)=f(t,t) to calculate I(t)
    B0_X = basis.get_basis(0.0)
    B0_Y = basis.get_aux_basis(0.0)

    for idx, (t_, w0, w1) in enumerate(zip(grid_t, W0, W1)):
        w1 = np.reshape(w1, (nb_path, 1))
        # interpolation of the volatility matrix is tedious
        idx_t = bracket(ts[1:], t_, throw_if_not_found=True)
        beta_t = betaxs[idx_t]
        volvol_t = volvolxs[idx_t]
        C_t = C[idx_t]
        Omega_t = Omega[idx_t]
        vartheta2 = np.dot(beta_t, beta_t) + volvol_t * volvol_t

        if measure_type == Measure.ANNUITY:
            ann_der0_mc = basis.annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, ccy=ccy, m=0)
            ann_der1_mc = basis.annuity(t=t_, ts_sw=ts_sw, x=x0, y=y0, ccy=ccy, m=1)
            d_loga_dx = divide_mc(ann_der1_mc, ann_der0_mc)
            adj_x_drift = prod_mc(np.dot(d_loga_dx, np.dot(C_t, np.transpose(C_t))), sigma0[:, 0] ** 2)
            adj_vol_drift = prod_mc(sigma0, np.dot(d_loga_dx, np.dot(C_t, beta_t)))
        elif measure_type == Measure.RISK_NEUTRAL:
            adj_x_drift = 0
            adj_vol_drift = 0
        elif measure_type == Measure.FORWARD:
            B_PX = basis.bond_coeffs(tau=T_fwd-t_)[0]
            CxCxB_P = np.dot(np.dot(C_t, np.transpose(C_t)), B_PX)
            adj_x_drift = -np.einsum('i,j->ji', CxCxB_P, sigma0[:, 0] ** 2)
            betaxCxB_P = np.dot(np.dot(B_PX, C_t), beta_t)
            adj_vol_drift = -sigma0 * betaxCxB_P

            # adj_x_drift = 0
            # adj_vol_drift = 0
        else:
            raise NotImplementedError

        Omega_t = np.tile(Omega_t, (nb_path, 1))
        # multiply by stoch vol driver
        Omega_t = prod_mc(Omega_t, sigma0[:, 0] * sigma0[:, 0])
        # if skew is modeled through DLN skew, not beta, formula for C and Omega are different
        if bxs is not None:
            ys = np.zeros((nb_path, params0.basis.get_nb_factors()))
            tenors = params0.basis.key_terms
            for idx_tenor, tenor in enumerate(tenors):
                ys[:, idx_tenor] = -1.0 / tenor * np.log(
                    params0.basis.bond(t=t_, T=t_ + tenor, x=x0, y=y0, ccy=params0.ccy))
            C_t = params0.calc_factor_vols_dln(yield_vols=A[idx_t], yields=ys, b_dln=bxs, nb_path=nb_path)
            # C_t_2 = params0.calc_factor_vols_dln2(t=t_, yield_vols=A[idx_t], b_dln=bxs)

            Omega_t = np.zeros((nb_path, params0.basis.get_nb_aux_factors()))
            for i, C_t_ith in enumerate(C_t):
                var_t = np.dot(C_t_ith, np.transpose(C_t_ith))
                Omega_t[i, :] = params0.basis.calc_Omega(var_t)

        # make mean 0
        # w0 = w0 - np.average(w0, axis=0)
        # w1 = w1 - np.average(w1, axis=0)

        I0 = I0 + dt * (x0.dot(B0_X) + y0.dot(B0_Y))
        y0 = y0 + dt * (y0.dot(np.transpose(D_Y)) + Omega_t)
        if bxs is not None:
            # if skew is modeled through DLN skew, not beta, formula for C and Omega are different
            # they become stochastic
            for i in range(nb_path):
                x0[i, :] = x0[i, :] + dt * x0[i, :].dot(np.transpose(D_X)) + w0[i, :].dot(np.transpose(C_t[i, :, :])) * \
                           sigma0[i, 0] + adj_x_drift * dt
        else:
            x0 = x0 + dt * x0.dot(np.transpose(D_X)) + prod_mc(w0.dot(np.transpose(C_t)), sigma0[:, 0]) + adj_x_drift * dt
            log_vol = log_vol + ((kappa1 * theta / sigma0) - (kappa1 - kappa2 * theta + 0.5 * vartheta2) - kappa2 * sigma0) * dt + (
                w0.dot(beta_t)).reshape(nb_path, 1) + volvol_t * w1 + adj_vol_drift * dt
        # sigma0 = sigma0 + (kappa1 + kappa2 * sigma0)*(theta-sigma0)*dt + sigma0*(beta_t * w0 + volvol_t * w1) + adj_vol_drift * dt
        sigma0 = np.exp(log_vol)
        if idx + 1 in idx_ttms:
            x0s.append(x0), y0s.append(y0), I0s.append(I0), sigma0s.append(sigma0)

    return x0s, y0s, I0s, sigma0s


def simulate_logsv_futures_MF2(params: MultiFactRateLogSvParams,
                               ttm,
                               t_start,
                               t_end,
                               basis_type,
                               f0: float = None,
                               W: List[np.ndarray] = None,
                               nb_path: int = 100000,
                               seed: int = None):
    """Simulate futures rate F_t under Q^T"""
    # basis_type = "NELSON-SIEGEL"
    sigma0 = params.sigma0
    theta = params.theta
    kappa1, kappa2 = params.kappa1, params.kappa2
    basis = params.basis
    ts = params.beta.ts
    A = params.A
    R = params.R
    C = params.C
    Omega = params.Omega
    betaxs = params.beta.xs
    volvolxs = params.volvol.xs
    ccy = params.ccy

    assert ts.shape[0] > 0 and ts[0] == 0.0
    nb_times = ts.size - 1

    # check inputs
    nb_factors = basis.get_nb_factors()
    assert A.ndim == 2 and A.shape[0] == nb_times and A.shape[1] == nb_factors
    assert betaxs.ndim == 2 and betaxs.shape[0] == nb_times and betaxs.shape[1] == nb_factors
    assert volvolxs.ndim == 1 and volvolxs.shape[0] == nb_times
    assert R.ndim == 2 and R.shape[0] == R.shape[1] == nb_factors
    assert Omega.ndim == 2 and Omega.shape[0] == nb_times and Omega.shape[1] == basis.get_nb_aux_factors()

    sigma0 = sigma0 * np.ones(nb_path)
    # pre-generate random numbers
    if seed is None:
        seed = 16
    np.random.seed(seed)  # fix seed
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm, nb_steps_per_year=720)
    if W is None:
        W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path, basis.get_nb_factors()))  # TODO: undo
        W1 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))  # TODO: undo

    log_vol = np.log(sigma0)
    Delta = t_end - t_start
    b1, b2, h1, h2, h0 = futures_conv_adj(t_start, basis_type, params, 0, Delta, ExpansionOrder.ZERO, True, grid_t)

    x0 = np.zeros((1, params.basis.get_nb_factors()))
    y0 = np.zeros((1, params.basis.get_nb_aux_factors()))
    sigma0 = params.sigma0 * np.ones((1, 1))
    expansion_order = ExpansionOrder.FIRST
    if f0 is None:
        f0 = calc_futures_rate(
            ccy=ccy, basis_type=basis_type, params=params,
            x0=x0, y0=y0, sigma0=sigma0,
            t0=0.0, ttm=ttm, t_start=t_start, t_end=t_end,
            Delta=Delta, expansion_order=expansion_order)[0][0]


    zeta0 = np.log(f0 + 1.0 / Delta)

    a, eta, _, _, _, beta, volvol = params.transform_QT_params(expiry=ttm, t_start=t_start, t_end=t_end, t_grid=grid_t)

    for idx, (t_, w0, w1) in enumerate(zip(grid_t, W0, W1)):
        # interpolation of the volatility matrix is tedious
        idx_t = bracket(ts[1:], t_, throw_if_not_found=True)
        beta_t = betaxs[idx_t]
        volvol_t = volvolxs[idx_t]
        # C_t = C[idx_t]
        # Omega_t = Omega[idx_t]
        vartheta2 = np.dot(beta_t, beta_t) + volvol_t * volvol_t

        # Omega_t = np.tile(Omega_t, (nb_path, 1))
        # multiply by stoch vol driver
        # Omega_t = prod_mc(Omega_t, sigma0[:, 0] * sigma0[:, 0])

        tau_start = t_start - t_
        tau_end = t_end - t_
        tau_exp = ttm - t_
        B_P_start = basis.bond_coeffs(tau_start)[0]
        B_P_end = basis.bond_coeffs(tau_end)[0]
        B_P_exp = basis.bond_coeffs(tau_exp)[0]
        h1_t = h1[idx]

        a_t = a[idx]
        eta_t = eta[idx]
        # kappa0_t = kappa0[idx]
        # kappa1_t = kappa1[idx]
        # kappa2_t = kappa2[idx]

        a0_t = a_t + beta_t*h1_t
        a1_t = volvol_t * h1_t
        # eta_t = np.dot(np.transpose(C_t), B_P_exp)
        # delta_t = np.dot(a0_t, eta_t)
        adj_vol_drift = np.dot(beta_t, eta_t)  # we put plus so that adjustment is additive to kappa2
        zeta0 = zeta0 + (-np.dot(a0_t, eta_t) - 0.5 * np.dot(a0_t, a0_t) - 0.5 * a1_t * a1_t) * sigma0 * sigma0 * dt + sigma0 * w0.dot(
            a0_t) + sigma0 * w1 * a1_t
        log_vol = log_vol + ((kappa1*theta/sigma0) - (kappa1 - kappa2*theta + 0.5 * vartheta2) - (kappa2 + adj_vol_drift) * sigma0) * dt + w0.dot(beta_t) + volvol_t * w1
        sigma0 = np.exp(log_vol)
        f0 = np.exp(zeta0) - 1.0 / Delta

    return f0


def simulate_logsv_futures_MF(ttm: float,
                              sigma0: np.ndarray,
                              theta: float,
                              kappa1: float,
                              kappa2: float,
                              ts: np.ndarray,
                              A: List[np.ndarray],  # for multi-factor model
                              R: List[np.ndarray],  # for multi-factor model
                              B: List[np.ndarray],  # for multi-factor model
                              C: List[np.ndarray],  # for multi-factor model
                              Omega: List[np.ndarray],  # for multi-factor model
                              h1: float,  # coefficient from futures conv adj
                              betaxs: np.ndarray,
                              volvolxs: np.ndarray,
                              basis: NelsonSiegel,
                              t_start: float,
                              t_end: float,
                              ccy: str,
                              f0: float = None,
                              nb_path: int = 100000,
                              seed: int = None,
                              W: List[np.ndarray] = None
                              ) -> np.ndarray:
    assert ts.shape[0] > 0 and ts[0] == 0.0
    nb_times = ts.size - 1

    # check inputs
    nb_factors = basis.get_nb_factors()
    assert A.ndim == 3 and A.shape[0] == nb_times and A.shape[1] == A.shape[2] == nb_factors
    assert betaxs.ndim == 2 and betaxs.shape[0] == nb_times and betaxs.shape[1] == nb_factors
    assert volvolxs.ndim == 1 and volvolxs.shape[0] == nb_times
    assert R.ndim == 2 and R.shape[0] == R.shape[1] == nb_factors and np.all(R.shape == B.shape)
    assert Omega.ndim == 2 and Omega.shape[0] == nb_times and Omega.shape[1] == basis.get_nb_aux_factors()

    if sigma0.shape[0] == 1:
        sigma0 = sigma0 * np.ones(nb_path)
    else:
        assert sigma0.shape[0] == nb_path

    # pre-generate random numbers
    if seed is None:
        seed = 16
    np.random.seed(seed)  # fix seed
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm)
    if W is None:
        W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path, basis.get_nb_factors()))  # TODO: undo
        W1 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))  # TODO: undo
    else:
        W0 = W[0] * np.sqrt(dt)
        W1 = W[1] * np.sqrt(dt)

    log_vol = np.log(sigma0)
    x0 = np.zeros((nb_path, basis.get_nb_factors(),))
    y0 = np.zeros((nb_path, basis.get_nb_aux_factors(),))
    # In case if we do not provide futures rate as an input, replace it by forward rate
    if f0 is None:
        f0 = basis.libor_rate(t=0, t_start=t_start, t_end=t_end, x=x0, y=y0, ccy=ccy)
    delta_frac = t_end - t_start
    zeta0 = np.log(f0 + 1.0/delta_frac)

    for idx, (t_, w0, w1) in enumerate(zip(grid_t, W0, W1)):
        # w1 = np.reshape(w1, (nb_path, 1))
        # interpolation of the volatility matrix is tedious
        idx_t = bracket(ts[1:], t_, throw_if_not_found=True)
        beta_t = betaxs[idx_t]
        volvol_t = volvolxs[idx_t]
        C_t = C[idx_t]
        vartheta2 = np.dot(beta_t, beta_t) + volvol_t * volvol_t
        tau_start = t_start - t_
        tau_end = t_end - t_
        tau_exp = ttm - t_
        B_P_start = basis.bond_coeffs(tau_start)[0]
        B_P_end = basis.bond_coeffs(tau_end)[0]
        B_P_exp = basis.bond_coeffs(tau_exp)[0]

        a_t = np.dot(np.transpose(C_t), B_P_end - B_P_start)
        eta_t = np.dot(np.transpose(C_t), B_P_exp)
        delta_t = np.dot(a_t, eta_t)
        adj_vol_drift = -np.dot(beta_t, eta_t) * sigma0
        zeta0 = zeta0 + (-delta_t - 0.5 * np.dot(a_t, a_t)) * sigma0 * sigma0 * dt + sigma0 * w0.dot(a_t)

        log_vol = log_vol + ((kappa1 * theta / sigma0) - (
                kappa1 - kappa2 * theta + 0.5 * vartheta2) - kappa2 * sigma0) * dt + w0.dot(
            beta_t) + volvol_t * w1 + adj_vol_drift * dt
        sigma0 = np.exp(log_vol)
        f0 = np.exp(zeta0) - 1.0 / delta_frac

    return f0


# @njit(cache=False, fastmath=True)  # TODO
def simulate_logsv_swap_approx_terminal(ttm: float,
                                        sigma0: np.ndarray,
                                        theta: float,
                                        kappa1: float,
                                        kappa2: float,
                                        ts: np.ndarray,
                                        axs: np.ndarray,
                                        betaxs: np.ndarray,
                                        volvolxs: np.ndarray,
                                        t0: float = 0.0,
                                        s0: np.ndarray = None,
                                        seed: float = None,
                                        nb_path: int = 100000
                                        ) -> np.ndarray:
    if sigma0.shape[0] == 1:
        sigma0 = sigma0 * np.ones(nb_path)
    else:
        assert sigma0.shape[0] == nb_path

    # pre-generate random numbers
    if seed is None:
        seed = 16
    # pre-generate random numbers
    np.random.seed(seed)  # fix seed
    nb_steps, dt, grid_t = set_time_grid(ttm=ttm)
    grid_t = t0 + grid_t
    W0 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))
    W1 = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))

    log_vol = np.log(sigma0)
    ts_sw = get_default_swap_term_structure(t0+ttm)
    if s0 is None:
        s0 = np.array([swap_rate(0, ts_sw=ts_sw, x=0, y=0)[0]])
    if s0.shape[0] == 1:
        s0 = s0 * np.ones(nb_path)
    swap_der = swap_rate(grid_t, ts_sw=ts_sw, x=0, y=0)[1]

    for t_, w0, w1 in zip(grid_t, W0, W1):
        a_t = pw_const(ts, axs, t_, False)
        beta_t = pw_const(ts, betaxs, t_, False)
        volvol_t = pw_const(ts, volvolxs, t_, False)
        vartheta2 = beta_t * beta_t + volvol_t * volvol_t

        ds_dx = swap_rate(t=t_, ts_sw=ts_sw, x=0, y=0)[1]
        ann_der0_mc = annuity(t=t_, ts_sw=ts_sw, x=0, y=0, m=0)
        ann_der1_mc = annuity(t=t_, ts_sw=ts_sw, x=0, y=0, m=1)
        d_loga_dx = ann_der1_mc / ann_der0_mc
        adj_vol_drift = a_t * beta_t * sigma0 * d_loga_dx

        s0 = s0 + ds_dx * a_t * sigma0 * w0
        log_vol = log_vol + ((kappa1 * theta / sigma0) - (
                kappa1 - kappa2 * theta + 0.5 * vartheta2) - kappa2 * sigma0) * dt + beta_t * w0 + volvol_t * w1 + adj_vol_drift * dt
        sigma0 = np.exp(log_vol)

    return s0


# @njit(cache=False, fastmath=True)  # TODO
def compute_mcapprox_payoff(ttm: float,
                            s_mc: np.ndarray,
                            strikes_ttm: np.ndarray,
                            optiontypes_ttm: np.ndarray):
    payoffsign = np.where(optiontypes_ttm == 'P', -1, 1).astype(float)
    option_prices = np.zeros_like(strikes_ttm)
    option_std = np.zeros_like(strikes_ttm)

    for idx, (strike, sign) in enumerate(zip(strikes_ttm, payoffsign)):
        option_prices[idx] = np.nanmean(np.maximum(sign * (s_mc - strike), 0))
        option_std[idx] = np.nanstd(np.maximum(sign * (s_mc - strike), 0))

    return option_prices, option_std / np.sqrt(s_mc.shape[0])


# @njit(cache=False, fastmath=True)  # TODO
def calculate_swap_rate_terminal(ttm: float,
                                 x0: np.ndarray,
                                 y0: np.ndarray,
                                 I0: np.ndarray,
                                 ts_sw: np.ndarray):
    # calculate swap rates using simulated states
    # ts_sw = get_default_swap_term_structure(ttm)
    s_mc = swap_rate(t=ttm, ts_sw=ts_sw, x=x0, y=y0, is_mc_mode=True)[0]
    ann_mc = annuity(t=ttm, ts_sw=ts_sw, x=x0, y=y0, m=0, is_mc_mode=True)
    numer = 1.0 / bond(t=0, T=ttm, x=0, y=0, m=0, is_mc_mode=True) * np.exp(I0)
    return ts_sw, s_mc, ann_mc, numer


def calculate_swap_rate(t: float,
                        ttm: float,
                        x0: np.ndarray,
                        y0: np.ndarray,
                        I0: np.ndarray,
                        ts_sw: np.ndarray = None):
    if ts_sw is None:
        ts_sw = get_default_swap_term_structure(ttm)
    # calculate swap rates using simulated states
    s_mc = swap_rate(t=t, ts_sw=ts_sw, x=x0, y=y0, is_mc_mode=True)[0]
    ann_mc = annuity(t=t, ts_sw=ts_sw, x=x0, y=y0, m=0, is_mc_mode=True)
    numer = 1.0 / bond(t=0, T=t, x=0, y=0, m=0, is_mc_mode=True) * np.exp(I0)
    return ts_sw, s_mc, ann_mc, numer


# @njit(cache=False, fastmath=True)  # TODO
def compute_mc_vars_payoff(ttm: float,
                           x0: np.ndarray,
                           y0: np.ndarray,
                           I0: np.ndarray,
                           strikes_ttm: np.ndarray,
                           optiontypes_ttm: np.ndarray,
                           is_annuity_measure: bool = False):
    ts_sw, s_mc, ann_mc, numer = calculate_swap_rate_terminal(ttm, x0, y0, I0)
    payoffsign = np.where(optiontypes_ttm == 'P', -1, 1).astype(float)
    option_prices = np.zeros_like(strikes_ttm)
    option_std = np.zeros_like(strikes_ttm)

    df = bond(0, ttm, 0, 0, 0, False)
    ann_crv = annuity(ttm, ts_sw, 0, 0, 0)
    s0_crv = swap_rate(ttm, ts_sw, 0, 0)[0]
    # numer = numer * df

    for idx, (strike, sign) in enumerate(zip(strikes_ttm, payoffsign)):
        if is_annuity_measure:
            option_prices[idx] = np.nanmean(np.maximum(sign * (s_mc - strike), 0))
            option_std[idx] = np.nanstd(np.maximum(sign * (s_mc - strike), 0))
        else:
            option_prices[idx] = np.nanmean(1. / numer * ann_mc * np.maximum(sign * (s_mc - strike), 0)) / ann_crv / df
            option_std[idx] = np.nanstd(1. / numer * ann_mc * np.maximum(sign * (s_mc - strike), 0)) / ann_crv / df

    return option_prices, option_std / np.sqrt(x0.shape[0])



"""
Implementation of log-normal stochastic volatility model
The lognormal sv model interface derives from ModelPricer
"""
# package
import numpy as np
import pandas as pd
from numba import njit
from numba.typed import List
from typing import Tuple, Optional
from scipy.optimize import minimize
from enum import Enum

# stochvolmodels
from stochvolmodels.utils.config import VariableType
import stochvolmodels.utils.mgf_pricer as mgfp
from stochvolmodels.utils.mc_payoffs import compute_mc_vars_payoff
from stochvolmodels.utils.funcs import to_flat_np_array, set_time_grid, timer, compute_histogram_data, set_seed

# stochvolmodels pricers
from stochvolmodels.pricers.logsv.logsv_params import LogSvParams
import stochvolmodels.pricers.logsv.affine_expansion as afe
from stochvolmodels.pricers.model_pricer import ModelPricer
from stochvolmodels.pricers.logsv.affine_expansion import ExpansionOrder
from stochvolmodels.pricers.logsv.vol_moments_ode import fit_model_vol_backbone_to_varswaps

# data
from stochvolmodels.data.option_chain import OptionChain
from stochvolmodels.data.test_option_chain import get_btc_test_chain_data


class LogsvModelCalibrationType(Enum):
    PARAMS4 = 1  # v0, theta, beta, volvol; kappa1, kappa2 are set externally
    PARAMS5 = 2  # v0, theta, kappa1, beta, volvol
    PARAMS6 = 3  # v0, theta, kappa1, kappa2, beta, volvol
    PARAMS_WITH_VARSWAP_FIT = 4  # beta, volvol; kappa1, kappa2 are set externally; term structure of varswap is fit


class ConstraintsType(Enum):
    UNCONSTRAINT = 1
    MMA_MARTINGALE = 2  # kappa_2 >= beta
    INVERSE_MARTINGALE = 3  # kappa_2 >= 2.0*beta
    MMA_MARTINGALE_MOMENT4 = 4  # kappa_2 >= beta &
    INVERSE_MARTINGALE_MOMENT4 = 5  # kappa_2 >= 2.0*beta


class CalibrationEngine(Enum):
    ANALYTIC = 1
    MC = 2


LOGSV_BTC_PARAMS = LogSvParams(sigma0=0.8376, theta=1.0413, kappa1=3.1844, kappa2=3.058, beta=0.1514, volvol=1.8458)


class LogSVPricer(ModelPricer):

    # @timer
    def price_chain(self,
                    option_chain: OptionChain,
                    params: LogSvParams,
                    is_spot_measure: bool = True,
                    **kwargs
                    ) -> List[np.ndarray]:
        """
        implementation of generic method price_chain using log sv wrapper
        """
        model_prices = logsv_chain_pricer(params=params,
                                          ttms=option_chain.ttms,
                                          forwards=option_chain.forwards,
                                          discfactors=option_chain.discfactors,
                                          strikes_ttms=option_chain.strikes_ttms,
                                          optiontypes_ttms=option_chain.optiontypes_ttms,
                                          is_spot_measure=is_spot_measure,
                                          **kwargs)
        return model_prices

    @timer
    def model_mc_price_chain(self,
                             option_chain: OptionChain,
                             params: LogSvParams,
                             is_spot_measure: bool = True,
                             variable_type: VariableType = VariableType.LOG_RETURN,
                             nb_path: int = 100000,
                             nb_steps: Optional[int] = None,
                             **kwargs
                             ) -> (List[np.ndarray], List[np.ndarray]):
        vol_backbone_etas = params.get_vol_backbone_etas(ttms=option_chain.ttms)
        return logsv_mc_chain_pricer(v0=params.sigma0,
                                     theta=params.theta,
                                     kappa1=params.kappa1,
                                     kappa2=params.kappa2,
                                     beta=params.beta,
                                     volvol=params.volvol,
                                     vol_backbone_etas=vol_backbone_etas,
                                     ttms=option_chain.ttms,
                                     forwards=option_chain.forwards,
                                     discfactors=option_chain.discfactors,
                                     strikes_ttms=option_chain.strikes_ttms,
                                     optiontypes_ttms=option_chain.optiontypes_ttms,
                                     is_spot_measure=is_spot_measure,
                                     variable_type=variable_type,
                                     nb_path=nb_path,
                                     nb_steps_per_year=nb_steps or int(360 * np.max(option_chain.ttms)) + 1)

    def set_vol_scaler(self, option_chain: OptionChain) -> float:
        """
        use chain vols to set the scaler
        """
        atm0 = option_chain.get_chain_atm_vols()[0]
        ttm0 = option_chain.ttms[0]
        return set_vol_scaler(sigma0=atm0, ttm=ttm0)

    @timer
    def calibrate_model_params_to_chain(self,
                                        option_chain: OptionChain,
                                        params0: LogSvParams,
                                        params_min: LogSvParams = LogSvParams(sigma0=0.1, theta=0.1, kappa1=0.25, kappa2=0.25, beta=-3.0, volvol=0.2),
                                        params_max: LogSvParams = LogSvParams(sigma0=1.5, theta=1.5, kappa1=10.0, kappa2=10.0, beta=3.0, volvol=3.0),
                                        is_vega_weighted: bool = True,
                                        is_unit_ttm_vega: bool = False,
                                        model_calibration_type: LogsvModelCalibrationType = LogsvModelCalibrationType.PARAMS5,
                                        constraints_type: ConstraintsType = ConstraintsType.UNCONSTRAINT,
                                        calibration_engine: CalibrationEngine = CalibrationEngine.ANALYTIC,
                                        nb_path: int = 100000,
                                        nb_steps: int = 360,
                                        seed: int = 10,
                                        **kwargs
                                        ) -> LogSvParams:
        """
        implementation of model calibration interface with nonlinear constraints
        """
        vol_scaler = self.set_vol_scaler(option_chain=option_chain)

        x, market_vols = option_chain.get_chain_data_as_xy()
        market_vols = to_flat_np_array(market_vols)  # market mid quotes

        if is_vega_weighted:
            vegas_ttms = option_chain.get_chain_vegas(is_unit_ttm_vega=is_unit_ttm_vega)
            # if is_unit_ttm_vega:
            vegas_ttms = [vegas_ttm/sum(vegas_ttm) for vegas_ttm in vegas_ttms]
            weights = to_flat_np_array(vegas_ttms)
        else:
            weights = np.ones_like(market_vols)

        if model_calibration_type == LogsvModelCalibrationType.PARAMS_WITH_VARSWAP_FIT:
            varswap_strikes = option_chain.get_slice_varswap_strikes(floor_with_atm_vols=True)
        else:
            varswap_strikes = None

        def parse_model_params(pars: np.ndarray) -> LogSvParams:
            if model_calibration_type == LogsvModelCalibrationType.PARAMS4:
                fit_params = LogSvParams(sigma0=pars[0],
                                         theta=pars[1],
                                         kappa1=params0.kappa1,
                                         kappa2=params0.kappa2,
                                         beta=pars[2],
                                         volvol=pars[3])
            elif model_calibration_type == LogsvModelCalibrationType.PARAMS5:
                fit_params = LogSvParams(sigma0=pars[0],
                                         theta=pars[1],
                                         kappa1=pars[2],
                                         kappa2=None,
                                         beta=pars[3],
                                         volvol=pars[4])
            elif model_calibration_type == LogsvModelCalibrationType.PARAMS_WITH_VARSWAP_FIT:
                fit_params = LogSvParams(sigma0=params0.sigma0,
                                         theta=params0.theta,
                                         kappa1=params0.kappa1,
                                         kappa2=params0.kappa2,
                                         beta=pars[0],
                                         volvol=pars[1])
                # set model backbone
                vol_backbone = fit_model_vol_backbone_to_varswaps(log_sv_params=fit_params,
                                                                  varswap_strikes=varswap_strikes)
                fit_params.set_vol_backbone(vol_backbone=vol_backbone)

            else:
                raise NotImplementedError(f"{model_calibration_type}")
            return fit_params

        if calibration_engine == CalibrationEngine.MC:
            W0s, W1s, dts = get_randoms_for_chain_valuation(ttms=option_chain.ttms, nb_path=nb_path, nb_steps_per_year=nb_steps, seed=seed)

        def objective(pars: np.ndarray, args: np.ndarray) -> float:
            params = parse_model_params(pars=pars)

            if calibration_engine == CalibrationEngine.ANALYTIC:
                model_vols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params, vol_scaler=vol_scaler)

            elif calibration_engine == CalibrationEngine.MC:
                option_prices_ttm, option_std_ttm = logsv_mc_chain_pricer_fixed_randoms(ttms=option_chain.ttms,
                                                                                        forwards=option_chain.forwards,
                                                                                        discfactors=option_chain.discfactors,
                                                                                        strikes_ttms=option_chain.strikes_ttms,
                                                                                        optiontypes_ttms=option_chain.optiontypes_ttms,
                                                                                        W0s=W0s,
                                                                                        W1s=W1s,
                                                                                        dts=dts,
                                                                                        v0=params.sigma0,
                                                                                        theta=params.theta,
                                                                                        kappa1=params.kappa1,
                                                                                        kappa2=params.kappa2,
                                                                                        beta=params.beta,
                                                                                        volvol=params.volvol,
                                                                                        vol_backbone_etas=params.get_vol_backbone_etas(ttms=option_chain.ttms))
                model_vols = option_chain.compute_model_ivols_from_chain_data(model_prices=option_prices_ttm)
                print(f"option_prices_ttm\n{option_prices_ttm}")
                print(f"model_vols\n{model_vols}")

            else:
                raise NotImplementedError(f"{calibration_engine}")

            resid = np.nansum(weights * np.square(to_flat_np_array(model_vols) - market_vols))
            return resid

        # parametric constraints
        def martingale_measure(pars: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            return params.kappa2 - params.beta

        def inverse_measure(pars: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            return params.kappa2 - 2.0 * params.beta

        def vol_4thmoment_finite(pars: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            kappa = params.kappa1 + params.kappa2 * params.theta
            return kappa - 1.5 * params.vartheta2

        # set initial params
        if model_calibration_type == LogsvModelCalibrationType.PARAMS4:
            # fit: v0, theta, beta, volvol; kappa1, kappa2 is given with params0
            p0 = np.array([params0.sigma0, params0.theta, params0.beta, params0.volvol])
            bounds = ((params_min.sigma0, params_max.sigma0),
                      (params_min.theta, params_max.theta),
                      (params_min.beta, params_max.beta),
                      (params_min.volvol, params_max.volvol))

        elif model_calibration_type == LogsvModelCalibrationType.PARAMS5:
            # fit: v0, theta, kappa1, beta, volvol; kappa2 is mapped as kappa1 / theta
            p0 = np.array([params0.sigma0, params0.theta, params0.kappa1, params0.beta, params0.volvol])
            bounds = ((params_min.sigma0, params_max.sigma0),
                      (params_min.theta, params_max.theta),
                      (params_min.kappa1, params_max.kappa1),
                      (params_min.beta, params_max.beta),
                      (params_min.volvol, params_max.volvol))

        elif model_calibration_type == LogsvModelCalibrationType.PARAMS_WITH_VARSWAP_FIT:
            # beta, volvol; kappa1, kappa2 are set externally; term structure of varswap is fit
            p0 = np.array([params0.beta, params0.volvol])
            bounds = ((params_min.beta, params_max.beta),
                      (params_min.volvol, params_max.volvol))
        else:
            raise NotImplementedError(f"{model_calibration_type}")

        options = {'disp': True, 'ftol': 1e-8}

        if constraints_type == ConstraintsType.UNCONSTRAINT:
            constraints = None
        elif constraints_type == ConstraintsType.MMA_MARTINGALE:
            constraints = ({'type': 'ineq', 'fun': martingale_measure})
        elif constraints_type == ConstraintsType.INVERSE_MARTINGALE:
            constraints = ({'type': 'ineq', 'fun': inverse_measure})
        elif constraints_type == ConstraintsType.MMA_MARTINGALE_MOMENT4:
            constraints = ({'type': 'ineq', 'fun': martingale_measure}, {'type': 'ineq', 'fun': vol_4thmoment_finite})
        elif constraints_type == ConstraintsType.INVERSE_MARTINGALE_MOMENT4:
            constraints = ({'type': 'ineq', 'fun': inverse_measure}, {'type': 'ineq', 'fun': vol_4thmoment_finite})
        else:
            raise NotImplementedError

        """
        match constraints_type:
            case ConstraintsType.UNCONSTRAINT:
                constraints = None
            case ConstraintsType.MMA_MARTINGALE:
                constraints = ({'type': 'ineq', 'fun': martingale_measure})
            case ConstraintsType.INVERSE_MARTINGALE:
                constraints = ({'type': 'ineq', 'fun': inverse_measure})
            case ConstraintsType.MMA_MARTINGALE_MOMENT4:
                constraints = ({'type': 'ineq', 'fun': martingale_measure}, {'type': 'ineq', 'fun': vol_4thmoment_finite})
            case ConstraintsType.INVERSE_MARTINGALE_MOMENT4:
                constraints = ({'type': 'ineq', 'fun': inverse_measure}, {'type': 'ineq', 'fun': vol_4thmoment_finite})
            case _:
                raise NotImplementedError
        """
        if constraints is not None:
            res = minimize(objective, p0, args=None, method='SLSQP', constraints=constraints, bounds=bounds, options=options)
        else:
            res = minimize(objective, p0, args=None, method='SLSQP', bounds=bounds, options=options)

        fit_params = parse_model_params(pars=res.x)

        return fit_params

    @timer
    def simulate_vol_paths(self,
                           params: LogSvParams,
                           brownians: np.ndarray = None,
                           ttm: float = 1.0,
                           nb_path: int = 100000,
                           is_spot_measure: bool = True,
                           nb_steps: int = None,
                           year_days: int = 360,
                           **kwargs
                           ) -> Tuple[np.ndarray, np.ndarray]:
        """
        simulate vols in dt_path grid
        """
        nb_steps = nb_steps or int(np.ceil(year_days * ttm))
        sigma_t, grid_t = simulate_vol_paths(ttm=ttm,
                                             v0=params.sigma0,
                                             theta=params.theta,
                                             kappa1=params.kappa1,
                                             kappa2=params.kappa2,
                                             beta=params.beta,
                                             volvol=params.volvol,
                                             nb_path=nb_path,
                                             is_spot_measure=is_spot_measure,
                                             nb_steps_per_year=nb_steps,
                                             brownians=brownians,
                                             **kwargs)
        return sigma_t, grid_t

    @timer
    def simulate_terminal_values(self,
                                 params: LogSvParams,
                                 ttm: float = 1.0,
                                 nb_path: int = 100000,
                                 is_spot_measure: bool = True,
                                 **kwargs
                                 ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        simulate terminal values
        """
        x0, sigma0, qvar0 = simulate_logsv_x_vol_terminal(ttm=ttm,
                                                          x0=np.zeros(nb_path),
                                                          sigma0=params.sigma0 * np.ones(nb_path),
                                                          qvar0=np.zeros(nb_path),
                                                          theta=params.theta,
                                                          kappa1=params.kappa1,
                                                          kappa2=params.kappa2,
                                                          beta=params.beta,
                                                          volvol=params.volvol,
                                                          nb_path=nb_path,
                                                          is_spot_measure=is_spot_measure)
        return x0, sigma0, qvar0

    @timer
    def logsv_pdfs(self,
                   params: LogSvParams,
                   ttm: float,
                   space_grid: np.ndarray,
                   is_stiff_solver: bool = False,
                   is_analytic: bool = False,
                   is_spot_measure: bool = True,
                   expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
                   variable_type: VariableType = VariableType.LOG_RETURN,
                   vol_scaler: float = None
                   ) -> np.ndarray:
        return logsv_pdfs(params=params,
                          ttm=ttm,
                          space_grid=space_grid,
                          is_stiff_solver=is_stiff_solver,
                          is_analytic=is_analytic,
                          is_spot_measure=is_spot_measure,
                          expansion_order=expansion_order,
                          variable_type=variable_type,
                          vol_scaler=vol_scaler
                          )


@njit(cache=False, fastmath=True)
def v0_implied(atm: float, beta: float, volvol: float, theta: float, kappa1: float, ttm: float):
    """
    approximation for short term model atm vol
    """
    beta2 = beta * beta
    volvol2 = volvol*volvol
    vartheta2 = beta2 + volvol2

    def simple():
        return atm - (beta2 + volvol2) * ttm / 4.0

    if np.abs(beta) > 1.0:  # cannot use approximation when beta is too high
        v0 = simple()
    else:
        numer = -24.0 - beta2 * ttm - 2.0 * vartheta2 * ttm + 12.0 * kappa1 * ttm + np.sqrt(np.square(24.0 + beta2 * ttm + 2.0 * vartheta2 * ttm - 12.0 * kappa1 * ttm) - 288.0 * beta * ttm * (
                -2.0 * atm + theta * kappa1 * ttm))
        denumer = 12.0 * beta * ttm
        if np.abs(denumer) > 1e-10:
            v0 = numer / denumer
        else:
            v0 = atm - (np.square(beta) + np.square(volvol)) * ttm / 4.0
    return v0


def set_vol_scaler(sigma0: float, ttm: float) -> float:
    return sigma0 * np.sqrt(np.minimum(np.min(ttm), 0.5 / 12.0))  # lower bound is two weeks


def logsv_chain_pricer(params: LogSvParams,
                       ttms: np.ndarray,
                       forwards: np.ndarray,
                       discfactors: np.ndarray,
                       strikes_ttms: List[np.ndarray],
                       optiontypes_ttms: List[np.ndarray],
                       is_stiff_solver: bool = False,
                       is_analytic: bool = False,
                       is_spot_measure: bool = True,
                       expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
                       variable_type: VariableType = VariableType.LOG_RETURN,
                       vol_scaler: float = None,
                       **kwargs
                       ) -> List[np.ndarray]:
    """
    wrapper to price option chain on variable_type
    to do: do numba implementation using numba consistent solver
    """
    # starting values
    if vol_scaler is None:  # for calibrations we fix one vol_scaler so the grid is not affected by v0
        vol_scaler = set_vol_scaler(sigma0=params.sigma0, ttm=np.min(ttms))

    phi_grid, psi_grid, theta_grid = mgfp.get_transform_var_grid(variable_type=variable_type,
                                                                 is_spot_measure=is_spot_measure,
                                                                 vol_scaler=vol_scaler)

    a_t0 = np.zeros((phi_grid.shape[0], afe.get_expansion_n(expansion_order)), dtype=np.complex128)
    ttm0 = 0.0

    # outputs as numpy lists
    model_prices_ttms = List()
    for ttm, forward, strikes_ttm, optiontypes_ttm, discfactor in zip(ttms, forwards, strikes_ttms, optiontypes_ttms, discfactors):
        vol_backbone_eta = params.get_vol_backbone_eta(tau=ttm)
        a_t0, log_mgf_grid = afe.compute_logsv_a_mgf_grid(ttm=ttm - ttm0,
                                                          phi_grid=phi_grid,
                                                          psi_grid=psi_grid,
                                                          theta_grid=theta_grid,
                                                          a_t0=a_t0,
                                                          is_analytic=is_analytic,
                                                          expansion_order=expansion_order,
                                                          is_stiff_solver=is_stiff_solver,
                                                          is_spot_measure=is_spot_measure,
                                                          **params.to_dict(),
                                                          vol_backbone_eta=vol_backbone_eta)

        if variable_type == VariableType.LOG_RETURN:
            option_prices = mgfp.vanilla_slice_pricer_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                                                    phi_grid=phi_grid,
                                                                    forward=forward,
                                                                    strikes=strikes_ttm,
                                                                    optiontypes=optiontypes_ttm,
                                                                    discfactor=discfactor,
                                                                    is_spot_measure=is_spot_measure)

        elif variable_type == VariableType.Q_VAR:
            option_prices = mgfp.slice_qvar_pricer_with_a_grid(log_mgf_grid=log_mgf_grid,
                                                               psi_grid=psi_grid,
                                                               ttm=ttm,
                                                               forward=forward,
                                                               strikes=strikes_ttm,
                                                               optiontypes=optiontypes_ttm,
                                                               discfactor=discfactor,
                                                               is_spot_measure=is_spot_measure)

        else:
            raise NotImplementedError

        model_prices_ttms.append(option_prices)
        ttm0 = ttm

    return model_prices_ttms


def logsv_pdfs(params: LogSvParams,
               ttm: float,
               space_grid: np.ndarray,
               is_stiff_solver: bool = False,
               is_analytic: bool = False,
               is_spot_measure: bool = True,
               expansion_order: ExpansionOrder = ExpansionOrder.SECOND,
               variable_type: VariableType = VariableType.LOG_RETURN,
               vol_scaler: float = None
               ) -> np.ndarray:
    """
    wrapper to compute model pdfs
    """
    # starting values
    if vol_scaler is None:  # for calibrations we fix one vol_scaler so the grid is not affected by v0
        vol_scaler = set_vol_scaler(sigma0=params.sigma0, ttm=ttm)

    phi_grid, psi_grid, theta_grid = mgfp.get_transform_var_grid(variable_type=variable_type,
                                                                 is_spot_measure=is_spot_measure,
                                                                 vol_scaler=vol_scaler)

    a_t0 = afe.get_init_conditions_a(phi_grid=phi_grid,
                                     psi_grid=psi_grid,
                                     theta_grid=theta_grid,
                                     n_terms=afe.get_expansion_n(expansion_order=expansion_order),
                                     variable_type=variable_type)

    # compute mgf
    a_t0, log_mgf_grid = afe.compute_logsv_a_mgf_grid(ttm=ttm,
                                                      phi_grid=phi_grid,
                                                      psi_grid=psi_grid,
                                                      theta_grid=theta_grid,
                                                      a_t0=a_t0,
                                                      is_analytic=is_analytic,
                                                      expansion_order=expansion_order,
                                                      is_stiff_solver=is_stiff_solver,
                                                      is_spot_measure=is_spot_measure,
                                                      **params.to_dict())

    # outputs as numpy lists
    if variable_type == VariableType.LOG_RETURN:
        transform_var_grid = phi_grid
        shift = 0.0
        scale = 1.0
    elif variable_type == VariableType.Q_VAR:  # scaled by ttm
        transform_var_grid = psi_grid
        shift = 0.0
        scale = 1.0 / ttm
    elif variable_type == VariableType.SIGMA:
        transform_var_grid = theta_grid
        shift = params.theta
        scale = 1.0
    else:
        raise NotImplementedError

    pdf = mgfp.pdf_with_mgf_grid(log_mgf_grid=log_mgf_grid,
                                 transform_var_grid=transform_var_grid,
                                 space_grid=space_grid,
                                 shift=shift,
                                 scale=scale)
    pdf = pdf / scale
    return pdf


@njit(cache=False, fastmath=True)
def logsv_mc_chain_pricer(ttms: np.ndarray,
                          forwards: np.ndarray,
                          discfactors: np.ndarray,
                          strikes_ttms: Tuple[np.ndarray,...],
                          optiontypes_ttms: Tuple[np.ndarray, ...],
                          v0: float,
                          theta: float,
                          kappa1: float,
                          kappa2: float,
                          beta: float,
                          volvol: float,
                          vol_backbone_etas: np.ndarray,
                          is_spot_measure: bool = True,
                          nb_path: int = 100000,
                          nb_steps_per_year: int = 360,
                          variable_type: VariableType = VariableType.LOG_RETURN
                          ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    # starting values
    x0 = np.zeros(nb_path)
    qvar0 = np.zeros(nb_path)
    sigma0 = v0*np.ones(nb_path)
    ttm0 = 0.0

    # outputs as numpy lists
    option_prices_ttm = List()
    option_std_ttm = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm, vol_backbone_eta in zip(ttms, forwards, discfactors,
                                                                                        strikes_ttms, optiontypes_ttms,
                                                                                        vol_backbone_etas):
        x0, sigma0, qvar0 = simulate_logsv_x_vol_terminal(ttm=ttm - ttm0,
                                                          x0=x0,
                                                          sigma0=sigma0,
                                                          qvar0=qvar0,
                                                          theta=theta,
                                                          kappa1=kappa1,
                                                          kappa2=kappa2,
                                                          beta=beta,
                                                          volvol=volvol,
                                                          vol_backbone_eta=vol_backbone_eta,
                                                          nb_path=nb_path,
                                                          nb_steps_per_year=nb_steps_per_year,
                                                          is_spot_measure=is_spot_measure)
        ttm0 = ttm
        option_prices, option_std = compute_mc_vars_payoff(x0=x0, sigma0=sigma0, qvar0=qvar0,
                                                           ttm=ttm,
                                                           forward=forward,
                                                           strikes_ttm=strikes_ttm,
                                                           optiontypes_ttm=optiontypes_ttm,
                                                           discfactor=discfactor,
                                                           variable_type=variable_type)
        option_prices_ttm.append(option_prices)
        option_std_ttm.append(option_std)

    return option_prices_ttm, option_std_ttm


@njit(cache=False, fastmath=False)
def simulate_vol_paths(ttm: float,
                       v0: float,
                       theta: float,
                       kappa1: float,
                       kappa2: float,
                       beta: float,
                       volvol: float,
                       is_spot_measure: bool = True,
                       nb_path: int = 100000,
                       nb_steps_per_year: int = 360,
                       brownians: np.ndarray = None,
                       **kwargs
                       ) -> Tuple[np.ndarray, np.ndarray]:
    """
    simulate vol paths on grid_t = [0.0, ttm]
    """
    sigma0 = v0 * np.ones(nb_path)

    nb_steps, dt, grid_t = set_time_grid(ttm=ttm, nb_steps_per_year=nb_steps_per_year)

    if brownians is None:
        brownians = np.sqrt(dt) * np.random.normal(0, 1, size=(nb_steps, nb_path))

    if is_spot_measure:
        alpha, adj = -1.0, 0.0
    else:
        alpha, adj = 1.0, beta

    vartheta2 = beta*beta + volvol*volvol
    vartheta = np.sqrt(vartheta2)
    vol_var = np.log(sigma0)
    sigma_t = np.zeros((nb_steps_per_year + 1, nb_path))  # sigma grid will increase to include the sigma_0 at t0 = 0
    sigma_t[0, :] = sigma0  # keep first value
    for t_, w1_ in enumerate(brownians):
        vol_var = vol_var + ((kappa1 * theta / sigma0 - kappa1) + kappa2*(theta-sigma0) + adj*sigma0 - 0.5*vartheta2) * dt + vartheta*w1_
        sigma0 = np.exp(vol_var)
        sigma_t[t_+1, :] = sigma0

    return sigma_t, grid_t


@njit(cache=False, fastmath=False)
def simulate_logsv_x_vol_terminal(ttm: float,
                                  x0:  np.ndarray,
                                  sigma0: np.ndarray,
                                  qvar0: np.ndarray,
                                  theta: float,
                                  kappa1: float,
                                  kappa2: float,
                                  beta: float,
                                  volvol: float,
                                  vol_backbone_eta: float = 1.0,
                                  is_spot_measure: bool = True,
                                  nb_path: int = 100000,
                                  nb_steps_per_year: int = 360,
                                  W0: Optional[np.ndarray] = None,
                                  W1: Optional[np.ndarray] = None,
                                  dt: Optional[float] = None
                                  ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    mc simulator for terminal values of log-return, vol sigma0, and qvar for log sv model
    """
    if x0.shape[0] == 1:  # initial value
        x0 = x0*np.zeros(nb_path)
    else:
        assert x0.shape[0] == nb_path

    if qvar0.shape[0] == 1:  # initial value
        qvar0 = np.zeros(nb_path)
    else:
        assert qvar0.shape[0] == nb_path

    if sigma0.shape[0] == 1:
        sigma0 = sigma0 * np.ones(nb_path)
    else:
        assert sigma0.shape[0] == nb_path
    if W0 is None and W1 is None:
        nb_steps1, dt, grid_t = set_time_grid(ttm=ttm, nb_steps_per_year=nb_steps_per_year)
        print(f"nb_steps1={nb_steps1}, dt={dt}")
        sdt = np.sqrt(dt)
        W0_ = sdt * np.random.normal(0, 1, size=(nb_steps1, nb_path))
        W1_ = sdt * np.random.normal(0, 1, size=(nb_steps1, nb_path))
    else:
        sdt = np.sqrt(dt)
        W0_ = sdt * W0
        W1_ = sdt * W1

    if is_spot_measure:
        alpha, adj = -1.0, 0.0
    else:
        alpha, adj = 1.0, beta*vol_backbone_eta  # ? vol_backbone_eta

    vartheta2 = beta*beta + volvol*volvol
    vol_backbone_eta2 = vol_backbone_eta * vol_backbone_eta
    vol_var = np.log(sigma0)
    for t_, (w0, w1) in enumerate(zip(W0_, W1_)):
        sigma0_2dt = vol_backbone_eta2 * sigma0 * sigma0 * dt
        x0 = x0 + alpha * 0.5 * sigma0_2dt + vol_backbone_eta * sigma0 * w0
        vol_var = vol_var + ((kappa1 * theta / sigma0 - kappa1) + kappa2*(theta-sigma0) + adj*sigma0 - 0.5*vartheta2) * dt + beta*w0+volvol*w1
        sigma0 = np.exp(vol_var)
        qvar0 = qvar0 + 0.5*(sigma0_2dt + vol_backbone_eta2 * sigma0 * sigma0 * dt)

    return x0, sigma0, qvar0


def get_randoms_for_chain_valuation(ttms: np.ndarray,
                                    nb_path: int = 100000,
                                    nb_steps_per_year: int = 360,
                                    seed: int = 10
                                    ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """
    we need to fix random normals for suesequent evaluation using mc slices
    outputs as numpy lists
    """
    #
    set_seed(seed)
    W0s = List()
    W1s = List()
    dts = List()
    ttm0 = 0.0
    for ttm in ttms:
        # qqq
        nb_steps_, dt, grid_t = set_time_grid(ttm=ttm - ttm0, nb_steps_per_year=nb_steps_per_year)
        W0s.append(np.random.normal(0, 1, size=(nb_steps_, nb_path)))
        W1s.append(np.random.normal(0, 1, size=(nb_steps_, nb_path)))
        dts.append(dt)
        ttm0 = ttm
    return W0s, W1s, dts


@njit(cache=False, fastmath=True)
def logsv_mc_chain_pricer_fixed_randoms(ttms: np.ndarray,
                                        forwards: np.ndarray,
                                        discfactors: np.ndarray,
                                        strikes_ttms: Tuple[np.ndarray,...],
                                        optiontypes_ttms: Tuple[np.ndarray, ...],
                                        W0s: Tuple[np.ndarray, ...],
                                        W1s: Tuple[np.ndarray, ...],
                                        dts: Tuple[np.ndarray, ...],
                                        v0: float,
                                        theta: float,
                                        kappa1: float,
                                        kappa2: float,
                                        beta: float,
                                        volvol: float,
                                        vol_backbone_etas: np.ndarray,
                                        is_spot_measure: bool = True,
                                        variable_type: VariableType = VariableType.LOG_RETURN
                                        ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    chain valuation using fixed randoms
    """
    # starting values
    nb_path = W0s[0].shape[1]
    x0 = np.zeros(nb_path)
    qvar0 = np.zeros(nb_path)
    sigma0 = v0*np.ones(nb_path)
    ttm0 = 0.0

    # outputs as numpy lists
    option_prices_ttm = List()
    option_std_ttm = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm, vol_backbone_eta, W0, W1, dt in zip(ttms, forwards, discfactors,
                                                                                                strikes_ttms, optiontypes_ttms,
                                                                                                vol_backbone_etas,
                                                                                                W0s, W1s, dts):
        x0, sigma0, qvar0 = simulate_logsv_x_vol_terminal(ttm=ttm - ttm0,
                                                          x0=x0,
                                                          sigma0=sigma0,
                                                          qvar0=qvar0,
                                                          theta=theta,
                                                          kappa1=kappa1,
                                                          kappa2=kappa2,
                                                          beta=beta,
                                                          volvol=volvol,
                                                          vol_backbone_eta=vol_backbone_eta,
                                                          nb_path=nb_path,
                                                          dt=dt,
                                                          is_spot_measure=is_spot_measure,
                                                          W0=W0,
                                                          W1=W1)
        ttm0 = ttm
        option_prices, option_std = compute_mc_vars_payoff(x0=x0, sigma0=sigma0, qvar0=qvar0,
                                                           ttm=ttm,
                                                           forward=forward,
                                                           strikes_ttm=strikes_ttm,
                                                           optiontypes_ttm=optiontypes_ttm,
                                                           discfactor=discfactor,
                                                           variable_type=variable_type)
        option_prices_ttm.append(option_prices)
        option_std_ttm.append(option_std)

    return option_prices_ttm, option_std_ttm


class UnitTests(Enum):
    CHAIN_PRICER = 1
    SLICE_PRICER = 2
    CALIBRATOR = 3
    MC_COMPARISION = 4
    MC_COMPARISION_QVAR = 5
    VOL_PATHS = 6
    TERMINAL_VALUES = 7
    MMA_INVERSE_MEASURE_VS_MC = 8


def run_unit_test(unit_test: UnitTests):

    import matplotlib.pyplot as plt
    import seaborn as sns
    import stochvolmodels.data.test_option_chain as chains

    if unit_test == UnitTests.CHAIN_PRICER:
        option_chain = get_btc_test_chain_data()
        logsv_pricer = LogSVPricer()
        model_prices = logsv_pricer.price_chain(option_chain=option_chain, params=LOGSV_BTC_PARAMS)
        print(model_prices)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain, params=LOGSV_BTC_PARAMS)

    if unit_test == UnitTests.SLICE_PRICER:
        ttm = 1.0
        forward = 1.0
        strikes = np.array([0.9, 1.0, 1.1])
        optiontypes = np.array(['P', 'C', 'C'])

        logsv_pricer = LogSVPricer()
        model_prices, vols = logsv_pricer.price_slice(params=LOGSV_BTC_PARAMS,
                                                      ttm=ttm,
                                                      forward=forward,
                                                      strikes=strikes,
                                                      optiontypes=optiontypes)
        print(model_prices)
        print(vols)

        for strike, optiontype in zip(strikes, optiontypes):
            model_price, vol = logsv_pricer.price_vanilla(params=LOGSV_BTC_PARAMS,
                                                          ttm=ttm,
                                                          forward=forward,
                                                          strike=strike,
                                                          optiontype=optiontype)
            print(f"{model_price}, {vol}")

    elif unit_test == UnitTests.CALIBRATOR:
        option_chain = get_btc_test_chain_data()
        logsv_pricer = LogSVPricer()
        fit_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=option_chain,
                                                                  params0=LOGSV_BTC_PARAMS)
        print(fit_params)
        logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain,
                                                 params=fit_params)

    elif unit_test == UnitTests.MC_COMPARISION:
        option_chain = get_btc_test_chain_data()
        logsv_pricer = LogSVPricer()
        logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain,
                                            params=LOGSV_BTC_PARAMS)

    elif unit_test == UnitTests.MC_COMPARISION_QVAR:
        from stochvolmodels.pricers.logsv.vol_moments_ode import compute_analytic_qvar
        logsv_pricer = LogSVPricer()
        ttms = {'1m': 1.0/12.0, '6m': 0.5}
        option_chain = chains.get_qv_options_test_chain_data()
        option_chain = OptionChain.get_slices_as_chain(option_chain, ids=list(ttms.keys()))
        forwards = np.array([compute_analytic_qvar(params=LOGSV_BTC_PARAMS, ttm=ttm, n_terms=4) for ttm in ttms.values()])
        print(f"QV forwards = {forwards}")

        option_chain.forwards = forwards  # replace forwards to imply BSM vols
        option_chain.strikes_ttms = List(forward * strikes_ttm for forward, strikes_ttm in zip(option_chain.forwards, option_chain.strikes_ttms))

        fig = logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain,
                                                  params=LOGSV_BTC_PARAMS,
                                                  variable_type=VariableType.Q_VAR)
    
    elif unit_test == UnitTests.VOL_PATHS:
        logsv_pricer = LogSVPricer()
        nb_path = 10
        sigma_t, grid_t = logsv_pricer.simulate_vol_paths(params=LOGSV_BTC_PARAMS,
                                                          nb_path=nb_path,
                                                          nb_steps=360)

        vol_paths = pd.DataFrame(sigma_t, index=grid_t, columns=[f"{x+1}" for x in range(nb_path)])
        print(vol_paths)

    elif unit_test == UnitTests.TERMINAL_VALUES:
        logsv_pricer = LogSVPricer()
        params = LOGSV_BTC_PARAMS
        xt, sigmat, qvart = logsv_pricer.simulate_terminal_values(params=params)
        hx = compute_histogram_data(data=xt, x_grid=params.get_x_grid(), name='Log-price')
        hsigmat = compute_histogram_data(data=sigmat, x_grid=params.get_sigma_grid(), name='Sigma')
        hqvar = compute_histogram_data(data=qvart, x_grid=params.get_qvar_grid(), name='Qvar')
        dfs = {'Log-price': hx, 'Sigma': hsigmat, 'Qvar': hqvar}

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(1, 3, figsize=(18, 10), tight_layout=True)
        for idx, (key, df) in enumerate(dfs.items()):
            axs[idx].fill_between(df.index, np.zeros_like(df.to_numpy()), df.to_numpy(),
                                  facecolor='lightblue', step='mid', alpha=0.8, lw=1.0)
            axs[idx].set_title(key)

    elif unit_test == UnitTests.MMA_INVERSE_MEASURE_VS_MC:
        option_chain = get_btc_test_chain_data()
        logsv_pricer = LogSVPricer()
        logsv_pricer.plot_comp_mma_inverse_options_with_mc(option_chain=option_chain,
                                                           params=LOGSV_BTC_PARAMS)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MC_COMPARISION_QVAR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
import pandas as pd


def compute_var_swap_strike(puts: pd.Series, calls: pd.Series, forward: float, ttm: float) -> float:
    joint_slice = pd.concat([puts.rename('puts'), calls.rename('calls')], axis=1).sort_index()
    strikes = joint_slice.index.to_numpy()
    otm = strikes < forward
    # dk = strikes[1:] - strikes[:-1]
    n = strikes.shape[0]
    dk = np.zeros(n)
    for idx in np.arange(n):
        if idx == 0:
            dk[idx] = strikes[1] - strikes[0]
        elif idx == n - 1:
            dk[idx] = strikes[idx] - strikes[idx-1]
        else:
            dk[idx] = 0.5*(strikes[idx+1] - strikes[idx-1])

    option_strip = np.where(otm, joint_slice['puts'].to_numpy(), joint_slice['calls'].to_numpy())
    var_swap_strike = 2.0 * np.nansum(dk*option_strip/np.square(strikes))
    atm_strike = strikes[otm == False][0]
    correction = np.square(forward/atm_strike - 1.0)
    var_swap_strike = (var_swap_strike-correction) / ttm
    var_swap_strike = np.sqrt(var_swap_strike)
    return var_swap_strike



import numpy as np
import pandas as pd
from typing import Dict, Union
from scipy.stats import norm
from scipy.optimize import curve_fit, brenth

ALPHA = 'alpha'
BETA = 'beta'
TOTAL_VOL = 'total_vol'
RHO = 'rho'


def get_alpha(f0: float,
              ttm: float,
              vol_atm: float,
              beta: float,
              rho: float,
              total_vol: float,
              shift: float):
    """
    Compute SABR parameter alpha from an ATM normal volatility.
    Alpha is determined as the root of a 3rd degree polynomial. Return a single
    scalar alpha.
    """
    f_pow_beta = np.power(f0+shift, beta)
    omega = -1.0 / 8 * beta * (2.0 - beta) / np.power(f0 + shift, 2.0 - 2.0 * beta)
    p = [1.0/3*ttm*f_pow_beta*omega,
         0.0,
         f_pow_beta + 1.0 / 24 * ttm * f_pow_beta * total_vol ** 2 * (2.0 - 3.0 * rho ** 2),
         -vol_atm]

    roots = np.roots(p)
    roots_real = np.extract(np.isreal(roots), np.real(roots))
    # Note: the double real roots case is not tested
    alpha_first_guess = vol_atm / np.power(f0+shift, beta)
    i_min = np.argmin(np.abs(roots_real - alpha_first_guess))
    return roots_real[i_min]


def calc_logsv_ivols(strikes: Union[float, np.ndarray],
                     f0: float,
                     ttm: float,
                     alpha: float,
                     rho: float,
                     total_vol: float,
                     beta: float,
                     shift: float,
                     is_alpha_atmvol: bool = False) -> Union[float, np.ndarray]:
    """
    returns SABR normal implied volatilities
    """
    assert f0 > 0
    if not np.all(strikes + shift > 0):
        raise ValueError('strike + shift must be positive')
    assert beta >= 0 and beta <= 1
    tol = 1e-6

    if is_alpha_atmvol:
        alpha = get_alpha(f0=f0, ttm=ttm, vol_atm=alpha, beta=beta, rho=rho, total_vol=total_vol, shift=shift)

    if isinstance(strikes, float):
        strikes = np.array([strikes])
    ivols = np.zeros_like(strikes)

    for idx, strike in enumerate(strikes):
        if (1.0 - beta) >= 1e-3:  # if beta is not 1
            zeta = total_vol / alpha * (np.power(strike + shift, 1.0 - beta) - np.power(f0 + shift, 1.0 - beta)) / (1.0 - beta)
            omega = -1.0 / 8 * beta * (2.0 - beta) / np.power(f0 + shift, 2.0 - 2.0 * beta)
            if np.fabs(strike - f0) > tol:
                m1 = (1.0 - beta) * (strike - f0) / (np.power(strike + shift, 1.0 - beta) - np.power(f0 + shift, 1.0 - beta))
            else:
                m1 = np.power(f0 + shift, beta)
        else:  # shifted lognormal case, beta = 1
            zeta = total_vol / alpha * np.log((strike + shift) / (f0 + shift))
            omega = -1.0 / 8
            if np.fabs(strike - f0) > tol:
                m1 = (strike - f0) / np.log((strike + shift) / (f0 + shift))
            else:
                m1 = np.power(f0 + shift, beta)
        y_zeta = np.log((rho + zeta + np.sqrt(1 + 2.0 * rho * zeta + zeta ** 2)) / (1.0 + rho))
        e_zeta = np.sqrt(1.0 + 2.0 * rho * zeta + zeta ** 2)
        if np.fabs(strike - f0) > tol:
            theta_zeta = total_vol ** 2 / 24.0 * (-1 + 3.0 * (rho + zeta - rho * e_zeta) / (y_zeta * e_zeta)) + \
                         omega * alpha ** 2 / 6.0 * (1.0 - rho ** 2 + ((rho + zeta) * e_zeta - rho) / y_zeta)
            zeta_by_yzeta = zeta / y_zeta
        else:
            theta_zeta = total_vol ** 2 / 24.0 * (2.0 - 3.0 * rho ** 2) + omega * alpha ** 2 / 3.0
            zeta_by_yzeta = 1.0
        mult = np.where(theta_zeta >= 0.0, 1.0 + theta_zeta * ttm, 1.0 / (1.0 - theta_zeta * ttm))
        ivols[idx] = alpha * m1 * zeta_by_yzeta * mult
    return ivols


def fit_logsv_ivols(strikes: np.ndarray,
                    mid_vols: np.ndarray,
                    f0: float,
                    beta: float,
                    shift: float,
                    ttm: float) -> Dict[str, float]:
    atm_fit_params = cals_logsv_parab_fit(strikes=strikes, mid_vols=mid_vols, f0=f0, beta=beta, shift=shift)
    bounds = ([0.001, 0.01, -0.999], [3.0*atm_fit_params[ALPHA], 5.0, 0.999])
    atm_fit_params[RHO] = np.maximum(-0.99, np.minimum(0.99, atm_fit_params[RHO])) if ~np.isnan(atm_fit_params[RHO]) else 0.0
    atm_fit_params[TOTAL_VOL] = np.maximum(0.01, np.minimum(3.0, atm_fit_params[TOTAL_VOL])) if ~np.isnan(atm_fit_params[TOTAL_VOL]) else 0.1
    p0 = np.array([atm_fit_params[ALPHA], atm_fit_params[TOTAL_VOL], atm_fit_params[RHO]])

    # f(x; params)
    ivol_func_0 = lambda log_strikes, alpha, total_vol, rho: calc_logsv_ivols(strikes=strikes,
                                                                              f0=f0,
                                                                              ttm=ttm,
                                                                              alpha=alpha,
                                                                              rho=rho,
                                                                              total_vol=total_vol,
                                                                              beta=beta,
                                                                              shift=shift)
    sigma = None
    popt, pcov = curve_fit(f=ivol_func_0,
                           xdata=strikes,
                           ydata=mid_vols,
                           bounds=bounds,
                           p0=p0,
                           sigma=sigma)
    fit_params = {ALPHA: popt[0], BETA: beta, TOTAL_VOL: popt[1], RHO: popt[2]}
    return fit_params



def cals_logsv_parab_fit(strikes: np.ndarray,
                         mid_vols: np.ndarray,
                         f0: float,
                         beta: float,
                         shift: float,
                         strike_step: float = 0.001
                         ) -> Dict[str, float]:
    """
    compute initial fit for alpha, total_vol and rho in SABR model
    """
    v0 = np.interp(x=f0, xp=strikes, fp=mid_vols)
    v0_m1 = np.interp(x=f0-strike_step, xp=strikes, fp=mid_vols)
    v0_p1 = np.interp(x=f0+strike_step, xp=strikes, fp=mid_vols)
    # derivs with respect K
    v1 = (v0_p1 - v0_m1) / (2.0 * strike_step)
    v2 = (v0_p1 - 2.0*v0 + v0_m1) / (strike_step**2)
    # derivs with respect to z
    v1 = v1 * (f0+shift)
    v2 = (f0+shift)**2 * v2 + v1

    alpha = v0 / np.power(f0+shift, beta)
    total_vol2 = 1.0 / np.power(f0 + shift, 2.0) * (v0**2*np.power(beta-1.0, 2.0) + 6.0*v1**2 + 6*v0*(v1-beta*v1+v2))
    total_vol = np.sqrt(total_vol2)
    rho = (v0 - beta * v0 + 2.0 * v1) / total_vol / (f0 + shift)
    sabr_params = {ALPHA: alpha, BETA: beta, TOTAL_VOL: total_vol, RHO: rho}
    return sabr_params


def get_delta_at_strikes(strikes: np.ndarray,
                         f0: float,
                         ttm: float,
                         sigma0: float,
                         rho: float,
                         total_vol: float,
                         beta: float,
                         shift: float,
                         optiontypes: np.ndarray = None
                         ) -> pd.Series:
    if optiontypes is None:
        optiontypes = np.repeat('C', strikes.size)
    st = np.sqrt(ttm)
    moneyness = f0 - strikes
    vol_st = st * calc_logsv_ivols(strikes=strikes, f0=f0, ttm=ttm, alpha=sigma0, rho=rho, total_vol=total_vol,
                          beta=beta, shift=shift)
    d = moneyness / vol_st
    deltas = np.where(optiontypes == "C", norm.cdf(d), norm.cdf(d)-1)

    return deltas


def infer_strikes_from_deltas(deltas: np.ndarray,
                              f0: float,
                              ttm: float,
                              sigma0: float,
                              rho: float,
                              total_vol: float,
                              beta: float,
                              shift: float
                              ) -> pd.Series:
    """
    givem
    """
    st = np.sqrt(ttm)
    def func(strike: float, given_delta: float) -> float:
        moneyness = f0-strike
        vol_st = st * calc_logsv_ivols(strikes=strike, f0=f0, ttm=ttm, alpha=sigma0, rho=rho, total_vol=total_vol,
                                       beta=beta, shift=shift)
        if given_delta >= 0.0:
            target = norm.ppf(given_delta)
        else:
            target = norm.ppf(1.0+given_delta)
        f = moneyness / vol_st - target
        return f

    imp_deltas = {}
    a = -shift + 0.0001
    b = 20*f0
    for idx, given_delta in enumerate(deltas):
        try:
            strike = brenth(f=func, a=a, b=b, args=(given_delta))  # , x0=forward
        except:
            print(f"can't find strike for delta={given_delta}, ttm={ttm}, forward={f0}")
            strike = f0
        imp_deltas[given_delta] = strike

    imp_deltas = pd.DataFrame.from_dict(imp_deltas, orient='index')
    return imp_deltas.iloc[:, 0]


